"""Enums for configuration element types used to construct/interact with OTX configuration objects."""
# Copyright (C) 2021-2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#


from __future__ import annotations

from enum import Enum, auto


class ElementCategory(Enum):
    """This Enum represents the categories of configuration elements that are known in OTX."""

    PRIMITIVES = auto()
    GROUPS = auto()
    RULES = auto()

    def __str__(self):
        """Retrieves the string representation of an instance of the Enum."""
        return self.name


class ConfigElementType(Enum):
    """This Enum represents the available elements to compose a configuration.

    Each instance holds a name, value and category representing a configuration element.
    """

    # Because this Enum takes both a value and a category, the auto() mechanism cannot be used to assign values. Hence,
    # they are assigned manually.
    INTEGER = 0, ElementCategory.PRIMITIVES
    FLOAT = 1, ElementCategory.PRIMITIVES
    BOOLEAN = 2, ElementCategory.PRIMITIVES
    FLOAT_SELECTABLE = 3, ElementCategory.PRIMITIVES
    SELECTABLE = 4, ElementCategory.PRIMITIVES
    PARAMETER_GROUP = 5, ElementCategory.GROUPS
    CONFIGURABLE_PARAMETERS = 6, ElementCategory.GROUPS
    RULE = 7, ElementCategory.RULES
    UI_RULES = 8, ElementCategory.RULES

    def __new__(cls, value: int, category: ElementCategory):  # pylint: disable=unused-argument
        """Creates a new instance of the Enum.

        The ConfigElementType Enum holds both a value and a category. In this method the `value` argument is parsed and
        assigned.
        """
        obj = object.__new__(cls)
        # Only the value is assigned here, since the _category_ attribute does not exists yet.
        obj._value_ = value
        return obj

    def __init__(self, value: int, category: ElementCategory):  # pylint: disable=unused-argument
        """Upon initialization, the Enum category is assigned."""
        # We cannot assign to _category_ in the __new__ method since it is not a valid attribute yet until the Enum is
        # initialized
        self._category_ = category

    @property
    def category(self) -> ElementCategory:
        """Returns the element category which the ConfigElementType belongs to.

        Categories are instances of the `otx.api.configuration.configuration_types.ElementCategory` Enum.
        """
        return self._category_

    def __str__(self):
        """Retrieves the string representation of an instance of the Enum."""
        return self.name



# Copyright (C) 2019-2021 Xiangtai Lee
# SPDX-License-Identifier: MIT
#
# Copyright (C) 2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#

import torch.nn as nn
import torch.nn.functional as F
from mmcv.cnn import ConvModule


class LocalAttentionModule(nn.Module):
    """Reference: https://github.com/lxtGH/GALD-DGCNet"""

    def __init__(self, num_channels, conv_cfg=None, norm_cfg=dict(type="BN")):
        super().__init__()

        self.num_channels = num_channels
        self.conv_cfg = conv_cfg
        self.norm_cfg = norm_cfg

        self.dwconv1 = ConvModule(
            in_channels=self.num_channels,
            out_channels=self.num_channels,
            kernel_size=3,
            stride=2,
            padding=1,
            groups=self.num_channels,
            conv_cfg=self.conv_cfg,
            norm_cfg=self.norm_cfg,
            act_cfg=dict(type="ReLU"),
        )
        self.dwconv2 = ConvModule(
            in_channels=self.num_channels,
            out_channels=self.num_channels,
            kernel_size=3,
            stride=2,
            padding=1,
            groups=self.num_channels,
            conv_cfg=self.conv_cfg,
            norm_cfg=self.norm_cfg,
            act_cfg=dict(type="ReLU"),
        )
        self.dwconv3 = ConvModule(
            in_channels=self.num_channels,
            out_channels=self.num_channels,
            kernel_size=3,
            stride=2,
            padding=1,
            groups=self.num_channels,
            conv_cfg=self.conv_cfg,
            norm_cfg=self.norm_cfg,
            act_cfg=dict(type="ReLU"),
        )
        self.sigmoid_spatial = nn.Sigmoid()

    def forward(self, x):
        _, _, h, w = x.size()

        y = self.dwconv1(x)
        y = self.dwconv2(y)
        y = self.dwconv3(y)
        y = F.interpolate(y, size=(h, w), mode="bilinear", align_corners=True)
        mask = self.sigmoid_spatial(y)

        out = x + x * mask

        return out



# Copyright (c) 2018-2020 Open-MMLab.
# SPDX-License-Identifier: Apache-2.0
#
# Copyright (c) 2021 DeLightCMU
# SPDX-License-Identifier: Apache-2.0
#
# Copyright (C) 2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#

"""Modified from: https://github.com/HRNet/Lite-HRNet"""


import mmcv
import torch
import torch.nn as nn
import torch.nn.functional as F
import torch.utils.checkpoint as cp
from mmcv.cnn import (
    ConvModule,
    build_conv_layer,
    build_norm_layer,
    constant_init,
    normal_init,
)
from mmcv.runner import BaseModule, load_checkpoint
from mmcv.utils.parrots_wrapper import _BatchNorm
from mmseg.models.backbones.resnet import BasicBlock, Bottleneck
from mmseg.models.builder import BACKBONES
from mmseg.utils import get_root_logger

from ..utils import (
    AsymmetricPositionAttentionModule,
    IterativeAggregator,
    LocalAttentionModule,
    channel_shuffle,
)


class NeighbourSupport(nn.Module):
    def __init__(self, channels, kernel_size=3, key_ratio=8, value_ratio=8, conv_cfg=None, norm_cfg=None):
        super().__init__()

        self.in_channels = channels
        self.key_channels = int(channels / key_ratio)
        self.value_channels = int(channels / value_ratio)
        self.kernel_size = kernel_size

        self.key = nn.Sequential(
            ConvModule(
                in_channels=self.in_channels,
                out_channels=self.key_channels,
                kernel_size=1,
                stride=1,
                conv_cfg=conv_cfg,
                norm_cfg=norm_cfg,
                act_cfg=dict(type="ReLU"),
            ),
            ConvModule(
                self.key_channels,
                self.key_channels,
                kernel_size=self.kernel_size,
                stride=1,
                padding=(self.kernel_size - 1) // 2,
                groups=self.key_channels,
                conv_cfg=conv_cfg,
                norm_cfg=norm_cfg,
                act_cfg=None,
            ),
            ConvModule(
                in_channels=self.key_channels,
                out_channels=self.kernel_size * self.kernel_size,
                kernel_size=1,
                stride=1,
                conv_cfg=conv_cfg,
                norm_cfg=norm_cfg,
                act_cfg=None,
            ),
        )
        self.value = nn.Sequential(
            ConvModule(
                in_channels=self.in_channels,
                out_channels=self.value_channels,
                kernel_size=1,
                stride=1,
                conv_cfg=conv_cfg,
                norm_cfg=norm_cfg,
                act_cfg=None,
            ),
            nn.Unfold(kernel_size=self.kernel_size, stride=1, padding=1),
        )
        self.out_conv = ConvModule(
            in_channels=self.value_channels,
            out_channels=self.in_channels,
            kernel_size=1,
            stride=1,
            conv_cfg=conv_cfg,
            norm_cfg=norm_cfg,
            act_cfg=None,
        )

    def forward(self, x):
        h, w = [int(_) for _ in x.size()[-2:]]

        key = self.key(x).view(-1, 1, self.kernel_size**2, h, w)
        weights = torch.softmax(key, dim=2)

        value = self.value(x).view(-1, self.value_channels, self.kernel_size**2, h, w)
        y = torch.sum(weights * value, dim=2)
        y = self.out_conv(y)

        out = x + y

        return out


class CrossResolutionWeighting(nn.Module):
    def __init__(
        self, channels, ratio=16, conv_cfg=None, norm_cfg=None, act_cfg=(dict(type="ReLU"), dict(type="Sigmoid"))
    ):
        super().__init__()

        if isinstance(act_cfg, dict):
            act_cfg = (act_cfg, act_cfg)
        assert len(act_cfg) == 2
        assert mmcv.is_tuple_of(act_cfg, dict)

        self.channels = channels
        total_channel = sum(channels)

        self.conv1 = ConvModule(
            in_channels=total_channel,
            out_channels=int(total_channel / ratio),
            kernel_size=1,
            stride=1,
            conv_cfg=conv_cfg,
            norm_cfg=norm_cfg,
            act_cfg=act_cfg[0],
        )
        self.conv2 = ConvModule(
            in_channels=int(total_channel / ratio),
            out_channels=total_channel,
            kernel_size=1,
            stride=1,
            conv_cfg=conv_cfg,
            norm_cfg=norm_cfg,
            act_cfg=act_cfg[1],
        )

    def forward(self, x):
        min_size = [int(_) for _ in x[-1].size()[-2:]]

        out = [F.adaptive_avg_pool2d(s, min_size) for s in x[:-1]] + [x[-1]]
        out = torch.cat(out, dim=1)
        out = self.conv1(out)
        out = self.conv2(out)
        out = torch.split(out, self.channels, dim=1)
        out = [s * F.interpolate(a, size=s.size()[-2:], mode="nearest") for s, a in zip(x, out)]

        return out


class SpatialWeighting(nn.Module):
    def __init__(self, channels, ratio=16, conv_cfg=None, act_cfg=(dict(type="ReLU"), dict(type="Sigmoid")), **kwargs):
        super().__init__()

        if isinstance(act_cfg, dict):
            act_cfg = (act_cfg, act_cfg)
        assert len(act_cfg) == 2
        assert mmcv.is_tuple_of(act_cfg, dict)

        self.global_avgpool = nn.AdaptiveAvgPool2d(1)
        self.conv1 = ConvModule(
            in_channels=channels,
            out_channels=int(channels / ratio),
            kernel_size=1,
            stride=1,
            conv_cfg=conv_cfg,
            act_cfg=act_cfg[0],
        )
        self.conv2 = ConvModule(
            in_channels=int(channels / ratio),
            out_channels=channels,
            kernel_size=1,
            stride=1,
            conv_cfg=conv_cfg,
            act_cfg=act_cfg[1],
        )

    def forward(self, x):
        out = self.global_avgpool(x)
        out = self.conv1(out)
        out = self.conv2(out)

        return x * out


class SpatialWeightingV2(nn.Module):
    """The original repo: https://github.com/DeLightCMU/PSA"""

    def __init__(self, channels, ratio=16, conv_cfg=None, norm_cfg=None, enable_norm=False, **kwargs):
        super().__init__()

        self.in_channels = channels
        self.internal_channels = int(channels / ratio)

        # channel-only branch
        self.v_channel = ConvModule(
            in_channels=self.in_channels,
            out_channels=self.internal_channels,
            kernel_size=1,
            stride=1,
            bias=False,
            conv_cfg=conv_cfg,
            norm_cfg=norm_cfg if enable_norm else None,
            act_cfg=None,
        )
        self.q_channel = ConvModule(
            in_channels=self.in_channels,
            out_channels=1,
            kernel_size=1,
            stride=1,
            bias=False,
            conv_cfg=conv_cfg,
            norm_cfg=norm_cfg if enable_norm else None,
            act_cfg=None,
        )
        self.out_channel = ConvModule(
            in_channels=self.internal_channels,
            out_channels=self.in_channels,
            kernel_size=1,
            stride=1,
            conv_cfg=conv_cfg,
            norm_cfg=norm_cfg,
            act_cfg=dict(type="Sigmoid"),
        )

        # spatial-only branch
        self.v_spatial = ConvModule(
            in_channels=self.in_channels,
            out_channels=self.internal_channels,
            kernel_size=1,
            stride=1,
            bias=False,
            conv_cfg=conv_cfg,
            norm_cfg=norm_cfg if enable_norm else None,
            act_cfg=None,
        )
        self.q_spatial = ConvModule(
            in_channels=self.in_channels,
            out_channels=self.internal_channels,
            kernel_size=1,
            stride=1,
            bias=False,
            conv_cfg=conv_cfg,
            norm_cfg=norm_cfg if enable_norm else None,
            act_cfg=None,
        )
        self.global_avgpool = nn.AdaptiveAvgPool2d(1)

    def _channel_weighting(self, x):
        h, w = [int(_) for _ in x.size()[-2:]]

        v = self.v_channel(x).view(-1, self.internal_channels, h * w)

        q = self.q_channel(x).view(-1, h * w, 1)
        q = torch.softmax(q, dim=1)

        y = torch.matmul(v, q)
        y = y.view(-1, self.internal_channels, 1, 1)
        y = self.out_channel(y)

        out = x * y

        return out

    def _spatial_weighting(self, x):
        h, w = [int(_) for _ in x.size()[-2:]]

        v = self.v_spatial(x)
        v = v.view(-1, self.internal_channels, h * w)

        q = self.q_spatial(x)
        q = self.global_avgpool(q)
        q = torch.softmax(q, dim=1)
        q = q.view(-1, 1, self.internal_channels)

        y = torch.matmul(q, v)
        y = y.view(-1, 1, h, w)
        y = torch.sigmoid(y)

        out = x * y

        return out

    def forward(self, x):
        y_channel = self._channel_weighting(x)
        y_spatial = self._spatial_weighting(x)
        out = y_channel + y_spatial

        return out


class ConditionalChannelWeighting(nn.Module):
    def __init__(
        self,
        in_channels,
        stride,
        reduce_ratio,
        conv_cfg=None,
        norm_cfg=dict(type="BN"),
        with_cp=False,
        dropout=None,
        weighting_module_version="v1",
        neighbour_weighting=False,
        dw_ksize=3,
    ):
        super().__init__()

        self.with_cp = with_cp
        self.stride = stride
        assert stride in [1, 2]

        spatial_weighting_module = SpatialWeighting if weighting_module_version == "v1" else SpatialWeightingV2
        branch_channels = [channel // 2 for channel in in_channels]

        self.cross_resolution_weighting = CrossResolutionWeighting(
            branch_channels, ratio=reduce_ratio, conv_cfg=conv_cfg, norm_cfg=norm_cfg
        )
        self.depthwise_convs = nn.ModuleList(
            [
                ConvModule(
                    channel,
                    channel,
                    kernel_size=dw_ksize,
                    stride=self.stride,
                    padding=dw_ksize // 2,
                    groups=channel,
                    conv_cfg=conv_cfg,
                    norm_cfg=norm_cfg,
                    act_cfg=None,
                )
                for channel in branch_channels
            ]
        )
        self.spatial_weighting = nn.ModuleList(
            [
                spatial_weighting_module(
                    channels=channel, ratio=4, conv_cfg=conv_cfg, norm_cfg=norm_cfg, enable_norm=True
                )
                for channel in branch_channels
            ]
        )

        self.neighbour_weighting = None
        if neighbour_weighting:
            self.neighbour_weighting = nn.ModuleList(
                [
                    NeighbourSupport(
                        channel, kernel_size=3, key_ratio=8, value_ratio=4, conv_cfg=conv_cfg, norm_cfg=norm_cfg
                    )
                    for channel in branch_channels
                ]
            )

        self.dropout = None
        if dropout is not None and dropout > 0:
            self.dropout = nn.ModuleList([nn.Dropout(p=dropout) for _ in branch_channels])

    def _inner_forward(self, x):
        x = [s.chunk(2, dim=1) for s in x]
        x1 = [s[0] for s in x]
        x2 = [s[1] for s in x]

        x2 = self.cross_resolution_weighting(x2)
        x2 = [dw(s) for s, dw in zip(x2, self.depthwise_convs)]

        if self.neighbour_weighting is not None:
            x2 = [nw(s) for s, nw in zip(x2, self.neighbour_weighting)]

        x2 = [sw(s) for s, sw in zip(x2, self.spatial_weighting)]

        if self.dropout is not None:
            x2 = [dropout(s) for s, dropout in zip(x2, self.dropout)]

        out = [torch.cat([s1, s2], dim=1) for s1, s2 in zip(x1, x2)]
        out = [channel_shuffle(s, 2) for s in out]

        return out

    def forward(self, x):
        if self.with_cp and x.requires_grad:
            out = cp.checkpoint(self._inner_forward, x)
        else:
            out = self._inner_forward(x)

        return out


class Stem(nn.Module):
    def __init__(
        self,
        in_channels,
        stem_channels,
        out_channels,
        expand_ratio,
        conv_cfg=None,
        norm_cfg=dict(type="BN"),
        with_cp=False,
        strides=(2, 2),
        extra_stride=False,
        input_norm=False,
    ):
        super().__init__()

        assert isinstance(strides, (tuple, list))
        assert len(strides) == 2

        self.in_channels = in_channels
        self.out_channels = out_channels
        self.conv_cfg = conv_cfg
        self.norm_cfg = norm_cfg
        self.with_cp = with_cp

        self.input_norm = None
        if input_norm:
            self.input_norm = nn.InstanceNorm2d(in_channels)

        self.conv1 = ConvModule(
            in_channels=in_channels,
            out_channels=stem_channels,
            kernel_size=3,
            stride=strides[0],
            padding=1,
            conv_cfg=self.conv_cfg,
            norm_cfg=self.norm_cfg,
            act_cfg=dict(type="ReLU"),
        )

        self.conv2 = None
        if extra_stride:
            self.conv2 = ConvModule(
                in_channels=stem_channels,
                out_channels=stem_channels,
                kernel_size=3,
                stride=2,
                padding=1,
                conv_cfg=self.conv_cfg,
                norm_cfg=self.norm_cfg,
                act_cfg=dict(type="ReLU"),
            )

        mid_channels = int(round(stem_channels * expand_ratio))
        branch_channels = stem_channels // 2
        if stem_channels == self.out_channels:
            inc_channels = self.out_channels - branch_channels
        else:
            inc_channels = self.out_channels - stem_channels

        self.branch1 = nn.Sequential(
            ConvModule(
                branch_channels,
                branch_channels,
                kernel_size=3,
                stride=strides[1],
                padding=1,
                groups=branch_channels,
                conv_cfg=conv_cfg,
                norm_cfg=norm_cfg,
                act_cfg=None,
            ),
            ConvModule(
                branch_channels,
                inc_channels,
                kernel_size=1,
                stride=1,
                padding=0,
                conv_cfg=conv_cfg,
                norm_cfg=norm_cfg,
                act_cfg=dict(type="ReLU"),
            ),
        )

        self.expand_conv = ConvModule(
            branch_channels,
            mid_channels,
            kernel_size=1,
            stride=1,
            padding=0,
            conv_cfg=conv_cfg,
            norm_cfg=norm_cfg,
            act_cfg=dict(type="ReLU"),
        )
        self.depthwise_conv = ConvModule(
            mid_channels,
            mid_channels,
            kernel_size=3,
            stride=strides[1],
            padding=1,
            groups=mid_channels,
            conv_cfg=conv_cfg,
            norm_cfg=norm_cfg,
            act_cfg=None,
        )
        self.linear_conv = ConvModule(
            mid_channels,
            branch_channels if stem_channels == self.out_channels else stem_channels,
            kernel_size=1,
            stride=1,
            padding=0,
            conv_cfg=conv_cfg,
            norm_cfg=norm_cfg,
            act_cfg=dict(type="ReLU"),
        )

    def _inner_forward(self, x):
        if self.input_norm is not None:
            x = self.input_norm(x)

        x = self.conv1(x)
        if self.conv2 is not None:
            x = self.conv2(x)

        x1, x2 = x.chunk(2, dim=1)

        x1 = self.branch1(x1)

        x2 = self.expand_conv(x2)
        x2 = self.depthwise_conv(x2)
        x2 = self.linear_conv(x2)

        out = torch.cat((x1, x2), dim=1)
        out = channel_shuffle(out, 2)

        return out

    def forward(self, x):
        if self.with_cp and x.requires_grad:
            out = cp.checkpoint(self._inner_forward, x)
        else:
            out = self._inner_forward(x)

        return out


class StemV2(nn.Module):
    def __init__(
        self,
        in_channels,
        stem_channels,
        out_channels,
        expand_ratio,
        conv_cfg=None,
        norm_cfg=dict(type="BN"),
        with_cp=False,
        num_stages=1,
        strides=(2, 2),
        extra_stride=False,
        input_norm=False,
    ):
        super().__init__()

        assert num_stages > 0
        assert isinstance(strides, (tuple, list))
        assert len(strides) == 1 + num_stages

        self.in_channels = in_channels
        self.out_channels = out_channels
        self.conv_cfg = conv_cfg
        self.norm_cfg = norm_cfg
        self.with_cp = with_cp
        self.num_stages = num_stages

        self.input_norm = None
        if input_norm:
            self.input_norm = nn.InstanceNorm2d(in_channels)

        self.conv1 = ConvModule(
            in_channels=in_channels,
            out_channels=stem_channels,
            kernel_size=3,
            stride=strides[0],
            padding=1,
            conv_cfg=self.conv_cfg,
            norm_cfg=self.norm_cfg,
            act_cfg=dict(type="ReLU"),
        )

        self.conv2 = None
        if extra_stride:
            self.conv2 = ConvModule(
                in_channels=stem_channels,
                out_channels=stem_channels,
                kernel_size=3,
                stride=2,
                padding=1,
                conv_cfg=self.conv_cfg,
                norm_cfg=self.norm_cfg,
                act_cfg=dict(type="ReLU"),
            )

        mid_channels = int(round(stem_channels * expand_ratio))
        internal_branch_channels = stem_channels // 2
        out_branch_channels = self.out_channels // 2

        self.branch1, self.branch2 = nn.ModuleList(), nn.ModuleList()
        for stage in range(1, num_stages + 1):
            self.branch1.append(
                nn.Sequential(
                    ConvModule(
                        internal_branch_channels,
                        internal_branch_channels,
                        kernel_size=3,
                        stride=strides[stage],
                        padding=1,
                        groups=internal_branch_channels,
                        conv_cfg=conv_cfg,
                        norm_cfg=norm_cfg,
                        act_cfg=None,
                    ),
                    ConvModule(
                        internal_branch_channels,
                        out_branch_channels if stage == num_stages else internal_branch_channels,
                        kernel_size=1,
                        stride=1,
                        padding=0,
                        conv_cfg=conv_cfg,
                        norm_cfg=norm_cfg,
                        act_cfg=dict(type="ReLU"),
                    ),
                )
            )

            self.branch2.append(
                nn.Sequential(
                    ConvModule(
                        internal_branch_channels,
                        mid_channels,
                        kernel_size=1,
                        stride=1,
                        padding=0,
                        conv_cfg=conv_cfg,
                        norm_cfg=norm_cfg,
                        act_cfg=dict(type="ReLU"),
                    ),
                    ConvModule(
                        mid_channels,
                        mid_channels,
                        kernel_size=3,
                        stride=strides[stage],
                        padding=1,
                        groups=mid_channels,
                        conv_cfg=conv_cfg,
                        norm_cfg=norm_cfg,
                        act_cfg=None,
                    ),
                    ConvModule(
                        mid_channels,
                        out_branch_channels if stage == num_stages else internal_branch_channels,
                        kernel_size=1,
                        stride=1,
                        padding=0,
                        conv_cfg=conv_cfg,
                        norm_cfg=norm_cfg,
                        act_cfg=dict(type="ReLU"),
                    ),
                )
            )

    def _inner_forward(self, x):
        if self.input_norm is not None:
            x = self.input_norm(x)

        y = self.conv1(x)
        if self.conv2 is not None:
            y = self.conv2(y)

        out_list = [y]
        for stage in range(self.num_stages):
            y1, y2 = y.chunk(2, dim=1)

            y1 = self.branch1[stage](y1)
            y2 = self.branch2[stage](y2)

            y = torch.cat((y1, y2), dim=1)
            y = channel_shuffle(y, 2)
            out_list.append(y)

        return out_list

    def forward(self, x):
        if self.with_cp and x.requires_grad:
            out = cp.checkpoint(self._inner_forward, x)
        else:
            out = self._inner_forward(x)

        return out


class ShuffleUnit(nn.Module):
    """InvertedResidual block for ShuffleNetV2 backbone.

    Args:
        in_channels (int): The input channels of the block.
        out_channels (int): The output channels of the block.
        stride (int): Stride of the 3x3 convolution layer. Default: 1
        conv_cfg (dict): Config dict for convolution layer.
            Default: None, which means using conv2d.
        norm_cfg (dict): Config dict for normalization layer.
            Default: dict(type='BN').
        act_cfg (dict): Config dict for activation layer.
            Default: dict(type='ReLU').
        with_cp (bool): Use checkpoint or not. Using checkpoint will save some
            memory while slowing down the training speed. Default: False.
    """

    def __init__(
        self,
        in_channels,
        out_channels,
        stride=1,
        conv_cfg=None,
        norm_cfg=dict(type="BN"),
        act_cfg=dict(type="ReLU"),
        with_cp=False,
    ):
        super().__init__()
        self.stride = stride
        self.with_cp = with_cp

        branch_features = out_channels // 2
        if self.stride == 1:
            assert in_channels == branch_features * 2, (
                f"in_channels ({in_channels}) should equal to "
                f"branch_features * 2 ({branch_features * 2}) "
                "when stride is 1"
            )

        if in_channels != branch_features * 2:
            assert self.stride != 1, (
                f"stride ({self.stride}) should not equal 1 when " f"in_channels != branch_features * 2"
            )

        if self.stride > 1:
            self.branch1 = nn.Sequential(
                ConvModule(
                    in_channels,
                    in_channels,
                    kernel_size=3,
                    stride=self.stride,
                    padding=1,
                    groups=in_channels,
                    conv_cfg=conv_cfg,
                    norm_cfg=norm_cfg,
                    act_cfg=None,
                ),
                ConvModule(
                    in_channels,
                    branch_features,
                    kernel_size=1,
                    stride=1,
                    padding=0,
                    conv_cfg=conv_cfg,
                    norm_cfg=norm_cfg,
                    act_cfg=act_cfg,
                ),
            )

        self.branch2 = nn.Sequential(
            ConvModule(
                in_channels if (self.stride > 1) else branch_features,
                branch_features,
                kernel_size=1,
                stride=1,
                padding=0,
                conv_cfg=conv_cfg,
                norm_cfg=norm_cfg,
                act_cfg=act_cfg,
            ),
            ConvModule(
                branch_features,
                branch_features,
                kernel_size=3,
                stride=self.stride,
                padding=1,
                groups=branch_features,
                conv_cfg=conv_cfg,
                norm_cfg=norm_cfg,
                act_cfg=None,
            ),
            ConvModule(
                branch_features,
                branch_features,
                kernel_size=1,
                stride=1,
                padding=0,
                conv_cfg=conv_cfg,
                norm_cfg=norm_cfg,
                act_cfg=act_cfg,
            ),
        )

    def _inner_forward(self, x):
        if self.stride > 1:
            out = torch.cat((self.branch1(x), self.branch2(x)), dim=1)
        else:
            x1, x2 = x.chunk(2, dim=1)
            out = torch.cat((x1, self.branch2(x2)), dim=1)

        out = channel_shuffle(out, 2)

        return out

    def forward(self, x):
        if self.with_cp and x.requires_grad:
            out = cp.checkpoint(self._inner_forward, x)
        else:
            out = self._inner_forward(x)

        return out


class LiteHRModule(nn.Module):
    def __init__(
        self,
        num_branches,
        num_blocks,
        in_channels,
        reduce_ratio,
        module_type,
        multiscale_output=False,
        with_fuse=True,
        conv_cfg=None,
        norm_cfg=dict(type="BN"),
        with_cp=False,
        dropout=None,
        weighting_module_version="v1",
        neighbour_weighting=False,
    ):
        super().__init__()
        self._check_branches(num_branches, in_channels)

        self.in_channels = in_channels
        self.num_branches = num_branches

        self.module_type = module_type
        self.multiscale_output = multiscale_output
        self.with_fuse = with_fuse
        self.norm_cfg = norm_cfg
        self.conv_cfg = conv_cfg
        self.with_cp = with_cp
        self.weighting_module_version = weighting_module_version
        self.neighbour_weighting = neighbour_weighting

        if self.module_type == "LITE":
            self.layers = self._make_weighting_blocks(num_blocks, reduce_ratio, dropout=dropout)
        elif self.module_type == "NAIVE":
            self.layers = self._make_naive_branches(num_branches, num_blocks)

        if self.with_fuse:
            self.fuse_layers = self._make_fuse_layers()
            self.relu = nn.ReLU()

    @staticmethod
    def _check_branches(num_branches, in_channels):
        """Check input to avoid ValueError."""

        if num_branches != len(in_channels):
            error_msg = f"NUM_BRANCHES({num_branches}) != NUM_INCHANNELS({len(in_channels)})"
            raise ValueError(error_msg)

    def _make_weighting_blocks(self, num_blocks, reduce_ratio, stride=1, dropout=None):
        layers = []
        for i in range(num_blocks):
            layers.append(
                ConditionalChannelWeighting(
                    self.in_channels,
                    stride=stride,
                    reduce_ratio=reduce_ratio,
                    conv_cfg=self.conv_cfg,
                    norm_cfg=self.norm_cfg,
                    with_cp=self.with_cp,
                    dropout=dropout,
                    weighting_module_version=self.weighting_module_version,
                    neighbour_weighting=self.neighbour_weighting,
                )
            )

        return nn.Sequential(*layers)

    def _make_one_branch(self, branch_index, num_blocks, stride=1):
        """Make one branch."""

        layers = [
            ShuffleUnit(
                self.in_channels[branch_index],
                self.in_channels[branch_index],
                stride=stride,
                conv_cfg=self.conv_cfg,
                norm_cfg=self.norm_cfg,
                act_cfg=dict(type="ReLU"),
                with_cp=self.with_cp,
            )
        ]
        for i in range(1, num_blocks):
            layers.append(
                ShuffleUnit(
                    self.in_channels[branch_index],
                    self.in_channels[branch_index],
                    stride=1,
                    conv_cfg=self.conv_cfg,
                    norm_cfg=self.norm_cfg,
                    act_cfg=dict(type="ReLU"),
                    with_cp=self.with_cp,
                )
            )

        return nn.Sequential(*layers)

    def _make_naive_branches(self, num_branches, num_blocks):
        """Make branches."""

        branches = []
        for i in range(num_branches):
            branches.append(self._make_one_branch(i, num_blocks))

        return nn.ModuleList(branches)

    def _make_fuse_layers(self):
        """Make fuse layer."""

        if self.num_branches == 1:
            return None

        num_branches = self.num_branches
        in_channels = self.in_channels
        num_out_branches = num_branches if self.multiscale_output else 1

        fuse_layers = []
        for i in range(num_out_branches):
            fuse_layer = []
            for j in range(num_branches):
                if j > i:
                    fuse_layer.append(
                        nn.Sequential(
                            build_conv_layer(
                                self.conv_cfg,
                                in_channels[j],
                                in_channels[i],
                                kernel_size=1,
                                stride=1,
                                padding=0,
                                bias=False,
                            ),
                            build_norm_layer(self.norm_cfg, in_channels[i])[1],
                        )
                    )
                elif j == i:
                    fuse_layer.append(None)
                else:
                    conv_downsamples = []
                    for k in range(i - j):
                        if k == i - j - 1:
                            conv_downsamples.append(
                                nn.Sequential(
                                    build_conv_layer(
                                        self.conv_cfg,
                                        in_channels[j],
                                        in_channels[j],
                                        kernel_size=3,
                                        stride=2,
                                        padding=1,
                                        groups=in_channels[j],
                                        bias=False,
                                    ),
                                    build_norm_layer(self.norm_cfg, in_channels[j])[1],
                                    build_conv_layer(
                                        self.conv_cfg,
                                        in_channels[j],
                                        in_channels[i],
                                        kernel_size=1,
                                        stride=1,
                                        padding=0,
                                        bias=False,
                                    ),
                                    build_norm_layer(self.norm_cfg, in_channels[i])[1],
                                )
                            )
                        else:
                            conv_downsamples.append(
                                nn.Sequential(
                                    build_conv_layer(
                                        self.conv_cfg,
                                        in_channels[j],
                                        in_channels[j],
                                        kernel_size=3,
                                        stride=2,
                                        padding=1,
                                        groups=in_channels[j],
                                        bias=False,
                                    ),
                                    build_norm_layer(self.norm_cfg, in_channels[j])[1],
                                    build_conv_layer(
                                        self.conv_cfg,
                                        in_channels[j],
                                        in_channels[j],
                                        kernel_size=1,
                                        stride=1,
                                        padding=0,
                                        bias=False,
                                    ),
                                    build_norm_layer(self.norm_cfg, in_channels[j])[1],
                                    nn.ReLU(inplace=True),
                                )
                            )
                    fuse_layer.append(nn.Sequential(*conv_downsamples))
            fuse_layers.append(nn.ModuleList(fuse_layer))

        return nn.ModuleList(fuse_layers)

    def forward(self, x):
        """Forward function."""

        if self.num_branches == 1:
            return [self.layers[0](x[0])]

        if self.module_type == "LITE":
            out = self.layers(x)
        elif self.module_type == "NAIVE":
            for i in range(self.num_branches):
                x[i] = self.layers[i](x[i])
            out = x

        if self.with_fuse:
            out_fuse = []
            for i in range(len(self.fuse_layers)):
                y = out[0] if i == 0 else self.fuse_layers[i][0](out[0])
                for j in range(self.num_branches):
                    if i == j:
                        fuse_y = out[j]
                    else:
                        fuse_y = self.fuse_layers[i][j](out[j])

                    if fuse_y.size()[-2:] != y.size()[-2:]:
                        fuse_y = F.interpolate(fuse_y, size=y.size()[-2:], mode="nearest")

                    y += fuse_y

                out_fuse.append(self.relu(y))

            out = out_fuse
        elif not self.multiscale_output:
            out = [out[0]]

        return out


@BACKBONES.register_module()
class LiteHRNet(BaseModule):
    """Lite-HRNet backbone.

    `High-Resolution Representations for Labeling Pixels and Regions
    <https://arxiv.org/abs/1904.04514>`_

    Args:
        extra (dict): detailed configuration for each stage of HRNet.
        in_channels (int): Number of input image channels. Default: 3.
        conv_cfg (dict): dictionary to construct and config conv layer.
        norm_cfg (dict): dictionary to construct and config norm layer.
        norm_eval (bool): Whether to set norm layers to eval mode, namely,
            freeze running stats (mean and var). Note: Effect on Batch Norm
            and its variants only. Default: False
        with_cp (bool): Use checkpoint or not. Using checkpoint will save some
            memory while slowing down the training speed.
        zero_init_residual (bool): whether to use zero init for last norm layer
            in resblocks to let them behave as identity.
    """

    def __init__(
        self,
        extra,
        in_channels=3,
        conv_cfg=None,
        norm_cfg=dict(type="BN"),
        norm_eval=False,
        with_cp=False,
        zero_init_residual=False,
        dropout=None,
        init_cfg=None,
    ):
        super().__init__(init_cfg=init_cfg)

        self.extra = extra
        self.conv_cfg = conv_cfg
        self.norm_cfg = norm_cfg
        self.norm_eval = norm_eval
        self.with_cp = with_cp
        self.zero_init_residual = zero_init_residual

        self.stem = Stem(
            in_channels,
            input_norm=self.extra["stem"]["input_norm"],
            stem_channels=self.extra["stem"]["stem_channels"],
            out_channels=self.extra["stem"]["out_channels"],
            expand_ratio=self.extra["stem"]["expand_ratio"],
            strides=self.extra["stem"]["strides"],
            extra_stride=self.extra["stem"]["extra_stride"],
            conv_cfg=self.conv_cfg,
            norm_cfg=self.norm_cfg,
        )

        self.enable_stem_pool = self.extra["stem"].get("out_pool", False)
        if self.enable_stem_pool:
            self.stem_pool = nn.AvgPool2d(kernel_size=3, stride=2)

        self.num_stages = self.extra["num_stages"]
        self.stages_spec = self.extra["stages_spec"]

        num_channels_last = [
            self.stem.out_channels,
        ]
        for i in range(self.num_stages):
            num_channels = self.stages_spec["num_channels"][i]
            num_channels = [num_channels[i] for i in range(len(num_channels))]

            setattr(self, "transition{}".format(i), self._make_transition_layer(num_channels_last, num_channels))

            stage, num_channels_last = self._make_stage(
                self.stages_spec, i, num_channels, multiscale_output=True, dropout=dropout
            )
            setattr(self, "stage{}".format(i), stage)

        self.out_modules = None
        if self.extra.get("out_modules") is not None:
            out_modules = []
            in_modules_channels, out_modules_channels = num_channels_last[-1], None
            if self.extra["out_modules"]["conv"]["enable"]:
                out_modules_channels = self.extra["out_modules"]["conv"]["channels"]
                out_modules.append(
                    ConvModule(
                        in_channels=in_modules_channels,
                        out_channels=out_modules_channels,
                        kernel_size=1,
                        stride=1,
                        padding=0,
                        conv_cfg=self.conv_cfg,
                        norm_cfg=self.norm_cfg,
                        act_cfg=dict(type="ReLU"),
                    )
                )
                in_modules_channels = out_modules_channels
            if self.extra["out_modules"]["position_att"]["enable"]:
                out_modules.append(
                    AsymmetricPositionAttentionModule(
                        in_channels=in_modules_channels,
                        key_channels=self.extra["out_modules"]["position_att"]["key_channels"],
                        value_channels=self.extra["out_modules"]["position_att"]["value_channels"],
                        psp_size=self.extra["out_modules"]["position_att"]["psp_size"],
                        conv_cfg=self.conv_cfg,
                        norm_cfg=self.norm_cfg,
                    )
                )
            if self.extra["out_modules"]["local_att"]["enable"]:
                out_modules.append(
                    LocalAttentionModule(
                        num_channels=in_modules_channels,
                        conv_cfg=self.conv_cfg,
                        norm_cfg=self.norm_cfg,
                    )
                )

            if len(out_modules) > 0:
                self.out_modules = nn.Sequential(*out_modules)
                num_channels_last.append(in_modules_channels)

        self.add_stem_features = self.extra.get("add_stem_features", False)
        if self.add_stem_features:
            self.stem_transition = nn.Sequential(
                ConvModule(
                    self.stem.out_channels,
                    self.stem.out_channels,
                    kernel_size=3,
                    stride=1,
                    padding=1,
                    groups=self.stem.out_channels,
                    conv_cfg=conv_cfg,
                    norm_cfg=norm_cfg,
                    act_cfg=None,
                ),
                ConvModule(
                    self.stem.out_channels,
                    num_channels_last[0],
                    kernel_size=1,
                    stride=1,
                    padding=0,
                    conv_cfg=conv_cfg,
                    norm_cfg=norm_cfg,
                    act_cfg=dict(type="ReLU"),
                ),
            )

            num_channels_last = [num_channels_last[0]] + num_channels_last

        self.with_aggregator = self.extra.get("out_aggregator") and self.extra["out_aggregator"]["enable"]
        if self.with_aggregator:
            self.aggregator = IterativeAggregator(
                in_channels=num_channels_last,
                min_channels=self.extra["out_aggregator"].get("min_channels", None),
                conv_cfg=self.conv_cfg,
                norm_cfg=self.norm_cfg,
            )

    def _make_transition_layer(self, num_channels_pre_layer, num_channels_cur_layer):
        """Make transition layer."""

        num_branches_cur = len(num_channels_cur_layer)
        num_branches_pre = len(num_channels_pre_layer)

        transition_layers = []
        for i in range(num_branches_cur):
            if i < num_branches_pre:
                if num_channels_cur_layer[i] != num_channels_pre_layer[i]:
                    transition_layers.append(
                        nn.Sequential(
                            build_conv_layer(
                                self.conv_cfg,
                                num_channels_pre_layer[i],
                                num_channels_pre_layer[i],
                                kernel_size=3,
                                stride=1,
                                padding=1,
                                groups=num_channels_pre_layer[i],
                                bias=False,
                            ),
                            build_norm_layer(self.norm_cfg, num_channels_pre_layer[i])[1],
                            build_conv_layer(
                                self.conv_cfg,
                                num_channels_pre_layer[i],
                                num_channels_cur_layer[i],
                                kernel_size=1,
                                stride=1,
                                padding=0,
                                bias=False,
                            ),
                            build_norm_layer(self.norm_cfg, num_channels_cur_layer[i])[1],
                            nn.ReLU(),
                        )
                    )
                else:
                    transition_layers.append(None)
            else:
                conv_downsamples = []
                for j in range(i + 1 - num_branches_pre):
                    in_channels = num_channels_pre_layer[-1]
                    out_channels = num_channels_cur_layer[i] if j == i - num_branches_pre else in_channels
                    conv_downsamples.append(
                        nn.Sequential(
                            build_conv_layer(
                                self.conv_cfg,
                                in_channels,
                                in_channels,
                                kernel_size=3,
                                stride=2,
                                padding=1,
                                groups=in_channels,
                                bias=False,
                            ),
                            build_norm_layer(self.norm_cfg, in_channels)[1],
                            build_conv_layer(
                                self.conv_cfg, in_channels, out_channels, kernel_size=1, stride=1, padding=0, bias=False
                            ),
                            build_norm_layer(self.norm_cfg, out_channels)[1],
                            nn.ReLU(),
                        )
                    )
                transition_layers.append(nn.Sequential(*conv_downsamples))

        return nn.ModuleList(transition_layers)

    def _make_stage(self, stages_spec, stage_index, in_channels, multiscale_output=True, dropout=None):
        num_modules = stages_spec["num_modules"][stage_index]
        num_branches = stages_spec["num_branches"][stage_index]
        num_blocks = stages_spec["num_blocks"][stage_index]
        reduce_ratio = stages_spec["reduce_ratios"][stage_index]
        with_fuse = stages_spec["with_fuse"][stage_index]
        module_type = stages_spec["module_type"][stage_index]
        weighting_module_version = stages_spec.get("weighting_module_version", "v1")
        neighbour_weighting = stages_spec.get("neighbour_weighting", False)

        modules = []
        for i in range(num_modules):
            # multi_scale_output is only used last module
            if not multiscale_output and i == num_modules - 1:
                reset_multiscale_output = False
            else:
                reset_multiscale_output = True

            modules.append(
                LiteHRModule(
                    num_branches,
                    num_blocks,
                    in_channels,
                    reduce_ratio,
                    module_type,
                    multiscale_output=reset_multiscale_output,
                    with_fuse=with_fuse,
                    conv_cfg=self.conv_cfg,
                    norm_cfg=self.norm_cfg,
                    with_cp=self.with_cp,
                    dropout=dropout,
                    weighting_module_version=weighting_module_version,
                    neighbour_weighting=neighbour_weighting,
                )
            )
            in_channels = modules[-1].in_channels

        return nn.Sequential(*modules), in_channels

    def init_weights(self, pretrained=None):
        """Initialize the weights in backbone.

        Args:
            pretrained (str, optional): Path to pre-trained weights.
                Defaults to None.
        """

        if isinstance(pretrained, str):
            logger = get_root_logger()
            load_checkpoint(self, pretrained, strict=False, logger=logger)
        elif pretrained is None:
            for m in self.modules():
                if isinstance(m, nn.Conv2d):
                    normal_init(m, std=0.001)
                elif isinstance(m, (_BatchNorm, nn.GroupNorm)):
                    constant_init(m, 1)

            if self.zero_init_residual:
                for m in self.modules():
                    if isinstance(m, Bottleneck):
                        constant_init(m.norm3, 0)
                    elif isinstance(m, BasicBlock):
                        constant_init(m.norm2, 0)
        else:
            raise TypeError("pretrained must be a str or None")

    def forward(self, x):
        """Forward function."""

        stem_outputs = self.stem(x)
        y_x2 = y_x4 = stem_outputs
        # y_x2, y_x4 = stem_outputs[-2:]
        y = y_x4

        if self.enable_stem_pool:
            y = self.stem_pool(y)

        y_list = [y]
        for i in range(self.num_stages):
            transition_modules = getattr(self, "transition{}".format(i))

            stage_inputs = []
            for j in range(self.stages_spec["num_branches"][i]):
                if transition_modules[j]:
                    if j >= len(y_list):
                        stage_inputs.append(transition_modules[j](y_list[-1]))
                    else:
                        stage_inputs.append(transition_modules[j](y_list[j]))
                else:
                    stage_inputs.append(y_list[j])

            stage_module = getattr(self, "stage{}".format(i))
            y_list = stage_module(stage_inputs)

        if self.out_modules is not None:
            y_list.append(self.out_modules(y_list[-1]))

        if self.add_stem_features:
            y_stem = self.stem_transition(y_x2)
            y_list = [y_stem] + y_list

        out = y_list
        if self.with_aggregator:
            out = self.aggregator(out)

        if self.extra.get("add_input", False):
            out = [x] + out

        return out

    def train(self, mode=True):
        """Convert the model into training mode."""

        super().train(mode)

        if mode and self.norm_eval:
            for m in self.modules():
                if isinstance(m, _BatchNorm):
                    m.eval()



import pytest
import torch

from otx.mpa.modules.models.classifiers.supcon_classifier import (
    ImageClassifier,
    SupConClassifier,
)
from tests.test_suite.e2e_test_system import e2e_pytest_unit


class TestSupConClassifier:
    @pytest.fixture(autouse=True)
    def setup(self, mocker) -> None:
        mocker.patch.object(ImageClassifier, "__init__", return_value=None)
        self.kwargs = dict(multilabel=False, hierarchical=False)
        self.supcon_classifier = SupConClassifier(backbone=None, **self.kwargs)

    @e2e_pytest_unit
    def test_forward_train(self, mocker):
        img = torch.rand(1, 3, 224, 224)
        gt_label = torch.rand(1, 1)
        self.supcon_classifier.extract_feat = mocker.MagicMock()
        self.supcon_classifier.head = mocker.MagicMock()
        losses = self.supcon_classifier.forward_train(img, gt_label, **self.kwargs)

        assert losses is not None



"""Segmentor config for semi-supervised learning."""

model = dict(
    type="MeanTeacherSegmentor",
    orig_type="OTXEncoderDecoder",
    unsup_weight=0.1,
    train_cfg=dict(mix_loss=dict(enable=False, weight=0.1)),
    test_cfg=dict(mode="whole", output_scale=5.0),
)
__norm_cfg = dict(type="BN", requires_grad=True)



# Copyright (C) 2023 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#

import os

import numpy as np
import pytest

from otx.algorithms.common.tasks import BaseTask
from otx.algorithms.detection.tasks import DetectionInferenceTask
from otx.api.configuration.helper import create
from otx.api.entities.datasets import DatasetEntity
from otx.api.entities.label import Domain, LabelEntity
from otx.api.entities.metrics import Performance, ScoreMetric
from otx.api.entities.model_template import TaskType, parse_model_template
from otx.api.entities.resultset import ResultSetEntity
from otx.api.usecases.evaluation.metrics_helper import MetricsHelper
from otx.api.usecases.tasks.interfaces.export_interface import ExportType
from tests.test_suite.e2e_test_system import e2e_pytest_unit
from tests.unit.algorithms.detection.test_helpers import (
    DEFAULT_DET_TEMPLATE_DIR,
    DEFAULT_ISEG_TEMPLATE_DIR,
    generate_det_dataset,
    init_environment,
)


class TestOTXDetectionTaskInference:
    @pytest.fixture(autouse=True)
    def setup(self, otx_model, tmp_dir_path) -> None:
        self.inference_task = dict()
        for template, task_type in zip(
            [DEFAULT_DET_TEMPLATE_DIR, DEFAULT_ISEG_TEMPLATE_DIR], (TaskType.DETECTION, TaskType.INSTANCE_SEGMENTATION)
        ):
            model_template = parse_model_template(os.path.join(template, "template.yaml"))
            hyper_parameters = create(model_template.hyper_parameters.data)
            task_env = init_environment(hyper_parameters, model_template, task_type=task_type)
            inference_task = DetectionInferenceTask(task_env, output_path=str(tmp_dir_path))
            self.inference_task[task_type] = inference_task
        self.output_path = str(tmp_dir_path)
        self.model = otx_model

    @e2e_pytest_unit
    @pytest.mark.parametrize(
        "task_type, domain",
        [(TaskType.DETECTION, Domain.DETECTION), (TaskType.INSTANCE_SEGMENTATION, Domain.INSTANCE_SEGMENTATION)],
    )
    def test_infer(self, task_type, domain, mocker):
        """Test infer method in DetectionInferenceTask."""
        dataset, labels = generate_det_dataset(task_type=task_type)
        if task_type == TaskType.DETECTION:
            fake_prediction = [[np.random.rand(1, 5)]]
        elif task_type == TaskType.INSTANCE_SEGMENTATION:
            sample = dataset[0]
            fake_prediction = [
                (
                    [np.random.rand(1, 5)],
                    [
                        [
                            {
                                "size": [sample.width, sample.height],
                                "counts": b"<4601OO11OO100O100O",
                            }
                        ]
                    ],
                )
            ]
        fake_feature_vectors = [np.zeros((1, 1, 1))]
        fake_saliency_maps = [None]
        fake_results = dict(
            outputs=dict(
                metric=1.0,
                detections=fake_prediction,
                feature_vectors=fake_feature_vectors,
                saliency_maps=fake_saliency_maps,
            )
        )
        mock_run_task = mocker.patch.object(BaseTask, "_run_task", return_value=fake_results)
        updated_dataset = self.inference_task[task_type].infer(dataset, None)
        mock_run_task.assert_called_once()
        for updated in updated_dataset:
            assert updated.annotation_scene.contains_any([LabelEntity(name=labels[0].name, domain=domain)])

    @e2e_pytest_unit
    @pytest.mark.parametrize(
        "task_type",
        [TaskType.DETECTION, TaskType.INSTANCE_SEGMENTATION],
    )
    def test_evaluate(self, task_type, mocker):
        """Test evaluate method in DetectionInferenceTask."""
        result_set = ResultSetEntity(
            model=self.model,
            ground_truth_dataset=DatasetEntity(),
            prediction_dataset=DatasetEntity(),
        )
        fake_metrics = mocker.patch("otx.api.usecases.evaluation.f_measure.FMeasure", autospec=True)
        fake_metrics.get_performance.return_value = Performance(
            score=ScoreMetric(name="fake", value=0.1), dashboard_metrics="mAP"
        )
        mocker.patch.object(MetricsHelper, "compute_f_measure", return_value=fake_metrics)
        self.inference_task[task_type].evaluate(result_set)
        assert result_set.performance.score.value == 0.1

    @e2e_pytest_unit
    @pytest.mark.parametrize(
        "task_type",
        [TaskType.DETECTION, TaskType.INSTANCE_SEGMENTATION],
    )
    def test_export(self, task_type, mocker):
        """Test export method in DetectionInferenceTask, expected RuntimeError without model file."""
        fake_output = {"outputs": {"bin": None, "xml": None}}
        mock_run_task = mocker.patch.object(BaseTask, "_run_task", return_value=fake_output)

        with pytest.raises(RuntimeError):
            self.inference_task[task_type].export(ExportType.OPENVINO, self.model)
            mock_run_task.assert_called_once()

    @e2e_pytest_unit
    @pytest.mark.parametrize(
        "task_type",
        [TaskType.DETECTION, TaskType.INSTANCE_SEGMENTATION],
    )
    def test_export_with_model_files(self, task_type, mocker):
        """Test export method in DetectionInferenceTask."""
        with open(f"{self.output_path}/model.xml", "wb") as f:
            f.write(b"foo")
        with open(f"{self.output_path}/model.bin", "wb") as f:
            f.write(b"bar")

        fake_output = {"outputs": {"bin": f"{self.output_path}/model.xml", "xml": f"{self.output_path}/model.bin"}}
        mock_run_task = mocker.patch.object(BaseTask, "_run_task", return_value=fake_output)
        self.inference_task[task_type].export(ExportType.OPENVINO, self.model)

        mock_run_task.assert_called_once()
        assert self.model.get_data("openvino.bin")
        assert self.model.get_data("openvino.xml")

    @e2e_pytest_unit
    @pytest.mark.parametrize(
        "task_type",
        [TaskType.DETECTION, TaskType.INSTANCE_SEGMENTATION],
    )
    def test_unload(self, task_type, mocker):
        """Test unload method in DetectionInferenceTask."""
        mock_cleanup = mocker.patch.object(BaseTask, "cleanup")
        self.inference_task[task_type].unload()
        mock_cleanup.assert_called_once()



"""Tests for Class-Incremental Learning for object detection with OTX CLI"""
# Copyright (C) 2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#
import copy
import os

import pytest
import torch

from otx.api.entities.model_template import parse_model_template
from otx.cli.registry import Registry
from tests.test_suite.e2e_test_system import e2e_pytest_component
from tests.test_suite.run_test_command import (
    get_template_dir,
    nncf_optimize_testing,
    otx_deploy_openvino_testing,
    otx_eval_deployment_testing,
    otx_eval_openvino_testing,
    otx_eval_testing,
    otx_explain_openvino_testing,
    otx_explain_testing,
    otx_export_testing,
    otx_hpo_testing,
    otx_resume_testing,
    otx_train_testing,
)

args = {
    "--train-data-roots": "tests/assets/car_tree_bug",
    "--val-data-roots": "tests/assets/car_tree_bug",
    "--test-data-roots": "tests/assets/car_tree_bug",
    "--input": "tests/assets/car_tree_bug/images/train",
    "train_params": ["params", "--learning_parameters.num_iters", "1", "--learning_parameters.batch_size", "2"],
}

# Training params for resume, num_iters*2
resume_params = [
    "params",
    "--learning_parameters.num_iters",
    "2",
    "--learning_parameters.batch_size",
    "4",
]

otx_dir = os.getcwd()


MULTI_GPU_UNAVAILABLE = torch.cuda.device_count() <= 1
default_template = parse_model_template(
    os.path.join("otx/algorithms/detection/configs", "instance_segmentation", "resnet50_maskrcnn", "template.yaml")
)
default_templates = [default_template]
default_templates_ids = [default_template.model_template_id]

templates = Registry("otx/algorithms/detection").filter(task_type="INSTANCE_SEGMENTATION").templates
templates_ids = [template.model_template_id for template in templates]


class TestInstanceSegmentationCLI:
    @e2e_pytest_component
    @pytest.mark.parametrize("template", templates, ids=templates_ids)
    def test_otx_train(self, template, tmp_dir_path):
        tmp_dir_path = tmp_dir_path / "ins_seg"
        otx_train_testing(template, tmp_dir_path, otx_dir, args)

    @e2e_pytest_component
    @pytest.mark.parametrize("template", default_templates, ids=default_templates_ids)
    def test_otx_resume(self, template, tmp_dir_path):
        tmp_dir_path = tmp_dir_path / "ins_seg/test_resume"
        otx_resume_testing(template, tmp_dir_path, otx_dir, args)
        template_work_dir = get_template_dir(template, tmp_dir_path)
        args1 = copy.deepcopy(args)
        args1["train_params"] = resume_params
        args1["--resume-from"] = f"{template_work_dir}/trained_for_resume_{template.model_template_id}/weights.pth"
        otx_resume_testing(template, tmp_dir_path, otx_dir, args1)

    @e2e_pytest_component
    @pytest.mark.parametrize("template", templates, ids=templates_ids)
    def test_otx_export(self, template, tmp_dir_path):
        tmp_dir_path = tmp_dir_path / "ins_seg"
        otx_export_testing(template, tmp_dir_path)

    @e2e_pytest_component
    @pytest.mark.parametrize("template", templates, ids=templates_ids)
    def test_otx_eval(self, template, tmp_dir_path):
        tmp_dir_path = tmp_dir_path / "ins_seg"
        otx_eval_testing(template, tmp_dir_path, otx_dir, args)

    @e2e_pytest_component
    @pytest.mark.parametrize("template", default_templates, ids=default_templates_ids)
    def test_otx_eval_openvino(self, template, tmp_dir_path):
        tmp_dir_path = tmp_dir_path / "ins_seg"
        otx_eval_openvino_testing(template, tmp_dir_path, otx_dir, args, threshold=1.0)

    @e2e_pytest_component
    @pytest.mark.parametrize("template", templates, ids=templates_ids)
    def test_otx_explain(self, template, tmp_dir_path):
        tmp_dir_path = tmp_dir_path / "ins_seg"
        otx_explain_testing(template, tmp_dir_path, otx_dir, args)

    @e2e_pytest_component
    @pytest.mark.parametrize("template", templates, ids=templates_ids)
    def test_otx_explain_openvino(self, template, tmp_dir_path):
        tmp_dir_path = tmp_dir_path / "ins_seg"
        otx_explain_openvino_testing(template, tmp_dir_path, otx_dir, args)

    @e2e_pytest_component
    @pytest.mark.parametrize("template", default_templates, ids=default_templates_ids)
    def test_otx_deploy_openvino(self, template, tmp_dir_path):
        tmp_dir_path = tmp_dir_path / "ins_seg"
        otx_deploy_openvino_testing(template, tmp_dir_path, otx_dir, args)

    @e2e_pytest_component
    @pytest.mark.parametrize("template", default_templates, ids=default_templates_ids)
    def test_otx_eval_deployment(self, template, tmp_dir_path):
        tmp_dir_path = tmp_dir_path / "ins_seg"
        otx_eval_deployment_testing(template, tmp_dir_path, otx_dir, args, threshold=1.0)

    @e2e_pytest_component
    @pytest.mark.parametrize("template", default_templates, ids=default_templates_ids)
    def test_otx_hpo(self, template, tmp_dir_path):
        tmp_dir_path = tmp_dir_path / "ins_seg/test_hpo"
        otx_hpo_testing(template, tmp_dir_path, otx_dir, args)

    @e2e_pytest_component
    @pytest.mark.parametrize("template", templates, ids=templates_ids)
    def test_nncf_optimize(self, template, tmp_dir_path):
        tmp_dir_path = tmp_dir_path / "ins_seg"
        if template.entrypoints.nncf is None:
            pytest.skip("nncf entrypoint is none")

        nncf_optimize_testing(template, tmp_dir_path, otx_dir, args)

    @e2e_pytest_component
    @pytest.mark.skip(reason="CVS-101246 Multi-GPU tests are stuck while CI is running")
    @pytest.mark.skipif(MULTI_GPU_UNAVAILABLE, reason="The number of gpu is insufficient")
    @pytest.mark.parametrize("template", default_templates, ids=default_templates_ids)
    def test_otx_multi_gpu_train(self, template, tmp_dir_path):
        tmp_dir_path = tmp_dir_path / "ins_seg/test_multi_gpu"
        args1 = copy.deepcopy(args)
        args1["--gpus"] = "0,1"
        otx_train_testing(template, tmp_dir_path, otx_dir, args1)



# Copyright (C) 2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#

import numpy as np
from mmcv.runner import wrap_fp16_model

from otx.mpa.deploy.utils import sync_batchnorm_2_batchnorm
from otx.mpa.exporter_mixin import ExporterMixin
from otx.mpa.registry import STAGES
from otx.mpa.utils.logger import get_logger

from .stage import ClsStage

logger = get_logger()


@STAGES.register_module()
class ClsExporter(ExporterMixin, ClsStage):
    def run(self, model_cfg, model_ckpt, data_cfg, **kwargs):  # noqa: C901
        """Run exporter stage"""

        precision = kwargs.get("precision", "FP32")
        model_builder = kwargs.get("model_builder", self.MODEL_BUILDER)

        def model_builder_helper(*args, **kwargs):
            model = model_builder(*args, **kwargs)
            # TODO: handle various input size
            model = sync_batchnorm_2_batchnorm(model, 2)

            if hasattr(model, "is_export"):
                model.is_export = True

            if precision == "FP16":
                wrap_fp16_model(model)
            elif precision == "INT8":
                from nncf.torch.nncf_network import NNCFNetwork

                assert isinstance(model, NNCFNetwork)

            return model

        kwargs["model_builder"] = model_builder_helper

        return super().run(model_cfg, model_ckpt, data_cfg, **kwargs)

    @staticmethod
    def naive_export(output_dir, model_builder, precision, cfg, model_name="model"):
        from mmcls.datasets.pipelines import Compose

        from ..deploy.apis import NaiveExporter

        def get_fake_data(cfg, orig_img_shape=(128, 128, 3)):
            pipeline = cfg.data.test.pipeline
            pipeline = Compose(pipeline)
            data = dict(img=np.zeros(orig_img_shape, dtype=np.uint8))
            data = pipeline(data)
            return data

        fake_data = get_fake_data(cfg)
        opset_version = 11

        NaiveExporter.export2openvino(
            output_dir,
            model_builder,
            cfg,
            fake_data,
            precision=precision,
            model_name=model_name,
            input_names=["data"],
            output_names=["logits"],
            opset_version=opset_version,
        )



"""Utils for Action recognition OpenVINO export task."""

# Copyright (C) 2022 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions
# and limitations under the License.

import glob
from functools import partial
from subprocess import DEVNULL, CalledProcessError, run  # nosec
from typing import List, Optional

import cv2
import numpy as np
import onnx
import torch
from mmaction.models import Recognizer3D
from mmcv.runner import BaseModule
from mmcv.utils import Config

try:
    from mmcv.onnx.symbolic import register_extra_symbolics
except ModuleNotFoundError as e:
    raise NotImplementedError("please update mmcv to version>=1.0.4") from e

from otx.algorithms.action.adapters.mmaction.models.detectors import AVAFastRCNN


def _convert_sync_batch_to_normal_batch(module: BaseModule):
    """Convert the syncBNs into normal BN3ds."""
    module_output = module
    if isinstance(module, torch.nn.SyncBatchNorm):
        module_output = torch.nn.BatchNorm3d(
            module.num_features, module.eps, module.momentum, module.affine, module.track_running_stats
        )
        if module.affine:
            module_output.weight.data = module.weight.data.clone().detach()
            module_output.bias.data = module.bias.data.clone().detach()
            # keep requires_grad unchanged
            module_output.weight.requires_grad = module.weight.requires_grad
            module_output.bias.requires_grad = module.bias.requires_grad
        module_output.running_mean = module.running_mean
        module_output.running_var = module.running_var
        module_output.num_batches_tracked = module.num_batches_tracked
    for name, child in module.named_children():
        module_output.add_module(name, _convert_sync_batch_to_normal_batch(child))
    del module
    return module_output


# pylint: disable=too-many-locals
def preprocess(
    clip_len: int,
    width: int,
    height: int,
    interval: int = 1,
    category: int = 1,
):
    """Pre-process for action detection structure.

    To get proper proposals from Faster-RCNN, the action detector needs real data
    """
    frames = []
    frame_dir = f"tests/assets/cvat_dataset/action_detection/train/{category}/images"
    mean = [123.675, 116.28, 103.53]
    std = [58.395, 57.12, 57.375]
    # TODO: allow only .jpg, .png exts
    rawframes = glob.glob(frame_dir + "/*")  # type: ignore[index]
    rawframes.sort()
    for rawframe in rawframes:
        frame = cv2.imread(rawframe)
        ori_h, ori_w, _ = frame.shape
        resized_frame = cv2.resize(frame, (width, height))
        resized_frame = cv2.cvtColor(resized_frame, cv2.COLOR_RGB2BGR)
        resized_frame = (resized_frame - mean) / std
        frames.append(resized_frame)
    np_frames = np.expand_dims(frames, axis=0)  # [1, T, H, W, C]
    np_frames = np_frames.transpose(0, 4, 1, 2, 3)  # [1, C, T, H, W]
    frame_inds = get_frame_inds(np_frames, clip_len, interval)
    np_frames = np_frames[:, :, frame_inds, :, :]
    torch_input = torch.Tensor(np_frames)
    one_meta = {
        "img_shape": torch.Tensor((height, width)),
        "ori_shape": (ori_h, ori_w),
        "pad_shape": (height, width),
        "filename": "demo_vid.png",
        "scale_factor": np.array([width / ori_w, height / ori_h, width / ori_w, height / ori_h]),
        "flip": False,
        "show_img": False,
        "flip_direction": None,
    }
    meta = [[one_meta]]
    return torch_input, meta


def get_frame_inds(np_frames: np.ndarray, clip_len: int, interval: int):
    """Get sampled index for given np_frames."""
    frame_len = np_frames.shape[2]
    ori_clip_len = clip_len * interval
    if frame_len > ori_clip_len - 1:
        start = (frame_len - ori_clip_len + 1) / 2
    else:
        start = 0
    frame_inds = np.arange(clip_len) * interval + int(start)
    frame_inds = np.clip(frame_inds, 0, frame_len - 1)
    frame_inds = frame_inds.astype(np.int)
    return frame_inds


# pylint: disable=too-many-arguments, too-many-locals, protected-access, no-member
def pytorch2onnx(
    model: BaseModule,
    input_shape: List[int],
    opset_version: int = 11,
    show: bool = False,
    output_file: Optional[str] = "tmp.onnx",
    is_localizer: bool = False,
):
    """Convert pytorch model to onnx model.

    Args:
        model (BaseModule): The pytorch model to be exported.
        input_shape (List[int]): The input tensor shape of the model.
        opset_version (int): Opset version of onnx used. Default: 11.
        show (bool): Determines whether to print the onnx model architecture.
            Default: False.
        output_file (str): Output onnx model name. Default: 'tmp.onnx'.
        is_localizer(bool): Determines this model is localizer or not
            Default: False.
    """
    model = _convert_sync_batch_to_normal_batch(model)
    input_tensor = torch.randn(input_shape)

    # onnx.export does not support kwargs
    if not isinstance(model, Recognizer3D):
        if isinstance(model, AVAFastRCNN):
            model.add_detector()
            model.patch_pools()
        input_tensor, meta = preprocess(input_shape[2], input_shape[3], input_shape[4])
        model.forward = partial(model.forward_infer, img_metas=meta)
        onnx_input = [input_tensor]
        input_names = ["data"]
        output_names = ["det_bboxes", "det_labels"]
    else:
        if hasattr(model, "forward_dummy"):
            # TODO Replace model.forward with model.onnx_export
            model.forward = model.forward_dummy
        elif hasattr(model, "_forward") and is_localizer:
            model.forward = model._forward
        else:
            raise NotImplementedError("Please implement the forward method for exporting.")
        onnx_input = input_tensor
        input_names = ["data"]
        output_names = ["logits"]

    model.cpu().eval()

    register_extra_symbolics(opset_version)
    torch.onnx.export(
        model,
        onnx_input,
        output_file,
        input_names=input_names,
        output_names=output_names,
        export_params=True,
        keep_initializers_as_inputs=True,
        verbose=show,
        opset_version=opset_version,
    )

    print(f"Successfully exported ONNX model: {output_file}")


def _get_mo_cmd():
    for mo_cmd in ("mo", "mo.py"):
        try:
            run([mo_cmd, "-h"], stdout=DEVNULL, stderr=DEVNULL, check=True)
            return mo_cmd
        except CalledProcessError:
            pass
    raise RuntimeError("OpenVINO Model Optimizer is not found or configured improperly")


# pylint: disable=no-member
def onnx2openvino(
    cfg: Config,
    onnx_model_path: Optional[str],
    output_dir_path: Optional[str],
    layout: str,
    input_shape: Optional[List[int]] = None,
    input_format: str = "bgr",
    precision: str = "FP32",
    pruning_transformation: bool = False,
):
    """Convert ONNX model into OpenVINO model."""
    cfg.model.pretrained = None
    cfg.data.test.test_mode = True

    onnx_model = onnx.load(onnx_model_path)
    _output_names = set(out.name for out in onnx_model.graph.output)
    # Clear names of the nodes that produce network's output blobs.
    for node in onnx_model.graph.node:
        if _output_names.intersection(node.output):
            node.ClearField("name")
    onnx.save(onnx_model, onnx_model_path)
    output_names = ",".join(_output_names)

    mo_cmd = _get_mo_cmd()

    normalize = None
    for pipeline in cfg.data.test.pipeline:
        if pipeline["type"] == "Normalize":
            normalize = pipeline
            break
    assert normalize, "Could not find normalize parameters in data pipeline"

    mean_values = normalize["mean"]
    scale_values = normalize["std"]
    command_line = [
        mo_cmd,
        f"--input_model={onnx_model_path}",
        f"--mean_values={mean_values}",
        f"--scale_values={scale_values}",
        f"--output_dir={output_dir_path}",
        f"--output={output_names}",
        f"--data_type={precision}",
        f"--source_layout={layout}",
    ]

    assert input_format.lower() in ["bgr", "rgb"]

    if input_shape is not None:
        command_line.append(f"--input_shape={input_shape}")
    if normalize["to_bgr"]:
        command_line.append("--reverse_input_channels")
    if pruning_transformation:
        command_line.extend(["--transform", "Pruning"])

    print(" ".join(command_line))

    run(command_line, check=True)


def export_model(
    model: BaseModule,
    config: Config,
    onnx_model_path: Optional[str] = None,
    output_dir_path: Optional[str] = None,
):
    """Export PyTorch model into OpenVINO model."""
    if isinstance(model, Recognizer3D):
        input_shape = [1, 1, 3, 8, 224, 224]
        layout = "??c???"
    else:
        input_shape = [1, 3, 32, 256, 256]
        layout = "bctwh"
    pytorch2onnx(model, input_shape=input_shape, output_file=onnx_model_path)
    onnx2openvino(config, onnx_model_path, output_dir_path, layout, input_shape)



"""Contrastive head to get contrastive loss with predictor head."""

# Copyright (C) 2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#
# pylint: disable=unused-argument
from typing import Any, Dict

import torch
import torch.nn.functional as F
from mmcls.models.builder import HEADS, build_neck
from torch import nn


@HEADS.register_module()
class ConstrastiveHead(nn.Module):
    """Head for contrastive learning.

    Args:
        predictor (dict): configurations for predictor.
        size_average (bool): whether averaging loss using batch size. Default value is True.
    """

    def __init__(self, predictor: Dict[str, Any], size_average: bool = True, **kwargs):
        super().__init__()
        self.predictor = build_neck(predictor)
        self.size_average = size_average

    def init_weights(self, init_linear: str = "normal"):
        """Initialize predictor weights.

        Args:
            init_linear (str): Option to initialize weights. Default: 'normal'
        """
        self.predictor.init_weights(init_linear=init_linear)

    def forward(self, inputs: torch.Tensor, targets: torch.Tensor):
        """Forward head.

        Args:
            inputs (Tensor): NxC input features.
            targets (Tensor): NxC target features.

        Returns:
            dict[str, Tensor]: A dictionary of loss components.
        """
        pred = self.predictor(inputs)
        pred_norm = F.normalize(pred, dim=1)
        target_norm = F.normalize(targets, dim=1)
        loss = 2 * inputs.size(0) - 2 * (pred_norm * target_norm).sum()
        if self.size_average:
            loss /= inputs.size(0)

        return dict(loss=loss)



# Copyright (C) 2021-2023 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#

import tempfile

from mmcv.runner import EpochBasedRunner
from mmcv.utils import get_logger
from torch.optim import SGD

from otx.algorithms.common.adapters.mmcv.utils import build_data_parallel
from otx.algorithms.segmentation.adapters.mmseg.nncf.hooks import (
    CustomstepLrUpdaterHook,
)
from tests.test_suite.e2e_test_system import e2e_pytest_unit
from tests.unit.algorithms.common.adapters.mmcv.nncf.test_helpers import (
    create_config,
    create_model,
)


class TestCustomstepLrUpdaterHook:
    @e2e_pytest_unit
    def test_get_regular_lr(self):
        mock_config = create_config()
        mock_model = create_model()

        with tempfile.TemporaryDirectory() as tempdir:
            runner = EpochBasedRunner(
                build_data_parallel(mock_model, mock_config),
                optimizer=SGD(mock_model.parameters(), lr=0.01, momentum=0.9),
                max_epochs=10,
                work_dir=tempdir,
                logger=get_logger("mmcv"),
            )
            runner.data_loader = range(10)

            hook = CustomstepLrUpdaterHook(1)
            hook.base_lr = [0.01, 0.1]
            hook.before_train_iter(runner)
            assert hook.get_regular_lr(runner) == [0.01, 0.1]
            runner._iter = 20
            assert hook.get_regular_lr(runner) == [0.01 * hook.gamma, 0.1 * hook.gamma]

            runner = EpochBasedRunner(
                build_data_parallel(mock_model, mock_config),
                optimizer={
                    "dummy": SGD(mock_model.parameters(), lr=0.01, momentum=0.9),
                },
                max_epochs=10,
                work_dir=tempdir,
                logger=get_logger("mmcv"),
            )
            runner.data_loader = range(10)

            hook = CustomstepLrUpdaterHook(1)
            hook.base_lr = {"dummy": [0.01, 0.1]}
            hook.before_train_iter(runner)
            assert hook.get_regular_lr(runner) == {"dummy": [0.01, 0.1]}
            runner._iter = 20
            assert hook.get_regular_lr(runner) == {"dummy": [0.01 * hook.gamma, 0.1 * hook.gamma]}

    @e2e_pytest_unit
    def test_get_fixed_lr(self):
        mock_config = create_config()
        mock_model = create_model()

        with tempfile.TemporaryDirectory() as tempdir:
            runner = EpochBasedRunner(
                build_data_parallel(mock_model, mock_config),
                optimizer=SGD(mock_model.parameters(), lr=0.01, momentum=0.9),
                max_epochs=10,
                work_dir=tempdir,
                logger=get_logger("mmcv"),
            )
            runner.data_loader = range(10)

            hook = CustomstepLrUpdaterHook(1, fixed="constant", fixed_iters=5, fixed_ratio=0.2)
            hook.base_lr = [1.0]
            hook.before_train_iter(runner)
            lrs = []
            for i in range(1, 25):
                lrs.extend(hook.get_fixed_lr(i, [1.0]))
            assert len(set(lrs)) == 1

            hook = CustomstepLrUpdaterHook(1, fixed="linear", fixed_iters=5, fixed_ratio=0.2)
            hook.base_lr = [1.0]
            hook.before_train_iter(runner)
            lrs = []
            for i in range(1, 25):
                lrs.extend(hook.get_fixed_lr(i, [1.0]))
            assert sorted(lrs) == lrs

            hook = CustomstepLrUpdaterHook(1, fixed="cos", fixed_iters=5, fixed_ratio=0.2)
            hook.base_lr = [1.0]
            hook.before_train_iter(runner)
            lrs = []
            for i in range(1, 25):
                lrs.extend(hook.get_fixed_lr(i, [1.0]))
            assert sorted(lrs) == lrs

    @e2e_pytest_unit
    def test_get_wramup_lr(self):
        mock_config = create_config()
        mock_model = create_model()

        with tempfile.TemporaryDirectory() as tempdir:
            runner = EpochBasedRunner(
                build_data_parallel(mock_model, mock_config),
                optimizer=SGD(mock_model.parameters(), lr=0.01, momentum=0.9),
                max_epochs=10,
                work_dir=tempdir,
                logger=get_logger("mmcv"),
            )
            runner.data_loader = range(10)

            hook = CustomstepLrUpdaterHook(1, warmup="constant", warmup_iters=5, warmup_ratio=0.2)
            hook.base_lr = [1.0]
            hook.before_train_iter(runner)
            lrs = []
            for i in range(1, 25):
                lrs.extend(hook.get_warmup_lr(i, [1.0]))
            assert len(set(lrs)) == 1

            hook = CustomstepLrUpdaterHook(1, warmup="linear", warmup_iters=5, warmup_ratio=0.2)
            hook.base_lr = [1.0]
            hook.before_train_iter(runner)
            lrs = []
            for i in range(1, 25):
                lrs.extend(hook.get_warmup_lr(i, [1.0]))
            assert sorted(lrs) == lrs



"""Utils for model io operations."""

# Copyright (C) 2021 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions
# and limitations under the License.

import json
import os
import re
import struct
import tempfile
from pathlib import Path
from typing import List, Optional, Tuple
from zipfile import ZipFile

import cv2
import numpy as np

from otx.api.entities.annotation import AnnotationSceneEntity, AnnotationSceneKind
from otx.api.entities.dataset_item import DatasetItemEntity
from otx.api.entities.datasets import DatasetEntity
from otx.api.entities.image import Image
from otx.api.entities.label_schema import LabelSchemaEntity
from otx.api.entities.model import (
    ModelConfiguration,
    ModelEntity,
    ModelOptimizationType,
)
from otx.api.serialization.label_mapper import LabelSchemaMapper
from otx.api.usecases.adapters.model_adapter import ModelAdapter
from otx.cli.utils.nncf import is_checkpoint_nncf


def save_model_data(model: ModelEntity, folder: str) -> None:
    """Saves model data to folder. Folder is created if it does not exist.

    Args:
        model (ModelEntity): The model to save.
        folder (str): Path to output folder.
    """

    os.makedirs(folder, exist_ok=True)
    for filename, model_adapter in model.model_adapters.items():
        with open(os.path.join(folder, filename), "wb") as write_file:
            write_file.write(model_adapter.data)


def read_binary(path: str) -> bytes:
    """Loads binary data stored at path.

    Args:
        path (str): A path where to load data from.

    Returns:
        bytes: Binary data.
    """

    with open(path, "rb") as read_file:
        return read_file.read()


def read_model(model_configuration: ModelConfiguration, path: str, train_dataset: DatasetEntity) -> ModelEntity:
    """Creates ModelEntity based on model_configuration and data stored at path.

    Args:
        model_configuration (ModelConfiguration): ModelConfiguration object.
        path (str): Path to the model data.
        train_dataset (DatasetEntity): DatasetEntity object.

    Returns:
        ModelEntity: ModelEntity object.
    """
    optimization_type = ModelOptimizationType.NONE

    model_adapter_keys = (
        "confidence_threshold",
        "image_threshold",
        "pixel_threshold",
        "min",
        "max",
        "config.json",
    )

    if path.endswith(".bin") or path.endswith(".xml"):
        # Openvino IR.
        model_adapters = {
            "openvino.xml": ModelAdapter(read_binary(path[:-4] + ".xml")),
            "openvino.bin": ModelAdapter(read_binary(path[:-4] + ".bin")),
        }
        for key in model_adapter_keys:
            full_path = os.path.join(os.path.dirname(path), key)
            if os.path.exists(full_path):
                model_adapters[key] = ModelAdapter(read_binary(full_path))
            else:
                model_adapters[key] = ModelAdapter(bytes())
    elif path.endswith(".pth"):
        # PyTorch
        model_adapters = {"weights.pth": ModelAdapter(read_binary(path))}

        if is_checkpoint_nncf(path):
            optimization_type = ModelOptimizationType.NNCF

        # Weights of auxiliary models
        for key in os.listdir(os.path.dirname(path)):
            if re.match(r"aux_model_[0-9]+\.pth", key):
                full_path = os.path.join(os.path.dirname(path), key)
                model_adapters[key] = ModelAdapter(read_binary(full_path))

    elif path.endswith(".zip"):
        # Deployed code.
        with tempfile.TemporaryDirectory() as temp_dir:
            with ZipFile(path) as myzip:
                myzip.extractall(temp_dir)

            model_path = os.path.join(temp_dir, "model", "model")
            model_adapters = {
                "openvino.xml": ModelAdapter(read_binary(model_path + ".xml")),
                "openvino.bin": ModelAdapter(read_binary(model_path + ".bin")),
            }

            config_path = os.path.join(temp_dir, "model", "config.json")
            with open(config_path, encoding="UTF-8") as f:
                model_parameters = json.load(f)["model_parameters"]
            model_adapters["config.json"] = ModelAdapter(read_binary(config_path))

            for key in model_adapter_keys:
                if key in model_parameters:
                    model_adapters[key] = ModelAdapter(struct.pack("f", model_parameters[key]))
    else:
        raise ValueError(f"Unknown file type: {path}")

    model = ModelEntity(
        configuration=model_configuration,
        model_adapters=model_adapters,
        train_dataset=train_dataset,
        optimization_type=optimization_type,
    )

    return model


def read_label_schema(path: str) -> LabelSchemaEntity:
    """Reads serialized LabelSchema and returns deserialized LabelSchema.

    Args:
        path (str): Path to model. It assmues that the `label_schema.json` is at the same location as the model.

    Returns:
        LabelSchemaEntity: Desetialized LabelSchemaEntity.
    """

    if any(path.endswith(extension) for extension in (".xml", ".bin", ".pth")):
        with open(os.path.join(os.path.dirname(path), "label_schema.json"), encoding="UTF-8") as read_file:
            serialized_label_schema = json.load(read_file)
    elif path.endswith(".zip"):
        with ZipFile(path) as read_zip_file:
            with read_zip_file.open(os.path.join("model", "config.json")) as read_file:
                serialized_label_schema = json.load(read_file)["model_parameters"]["labels"]
    return LabelSchemaMapper().backward(serialized_label_schema)


def get_image_files(root_dir: str) -> Optional[List[Tuple[str, str]]]:
    """Recursively get all image file paths from given root_dir."""
    img_data_formats = (
        ".jpg",
        ".JPG",
        ".jpeg",
        ".JPEG",
        ".gif",
        ".GIF",
        ".bmp",
        ".BMP",
        ".tif",
        ".TIF",
        ".tiff",
        ".TIFF",
        ".png",
        ".PNG",
    )
    # single image path
    if root_dir.endswith(img_data_formats):
        return [("./", root_dir)]

    img_files = []
    for root, _, _ in os.walk(root_dir):
        for format_ in img_data_formats:
            img_files.extend([(root, file.name) for file in Path(root).glob(f"*{format_}")])
    return img_files if img_files else None


def save_saliency_output(
    process_saliency_maps: bool,
    img: np.array,
    saliency_map: np.array,
    save_dir: str,
    fname: str,
    weight: float = 0.3,
) -> None:
    """Saves processed saliency map (with image overlay) or raw saliency map."""
    if process_saliency_maps:
        # Saves processed saliency map
        overlay = img * weight + saliency_map * (1 - weight)
        overlay[overlay > 255] = 255
        overlay = overlay.astype(np.uint8)

        cv2.imwrite(f"{os.path.join(save_dir, fname)}_saliency_map.png", saliency_map)
        cv2.imwrite(f"{os.path.join(save_dir, fname)}_overlay_img.png", overlay)
    else:
        # Saves raw, low-resolution saliency map
        cv2.imwrite(f"{os.path.join(save_dir, fname)}_saliency_map.tiff", saliency_map)


def get_explain_dataset_from_filelist(image_files: list):
    """Get explain dataset with empty annotation."""
    empty_annotation = AnnotationSceneEntity(annotations=[], kind=AnnotationSceneKind.PREDICTION)
    items = []
    for root_dir, filename in image_files:
        frame = cv2.imread(os.path.join(root_dir, filename))
        item = DatasetItemEntity(
            media=Image(cv2.cvtColor(frame, cv2.COLOR_BGR2RGB)),
            annotation_scene=empty_annotation,
        )
        items.append(item)
    explain_dataset = DatasetEntity(items=items)
    return explain_dataset



# Copyright (C) 2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#

from mmcv.runner import HOOKS, Hook, get_dist_info
from torch.utils.data import DataLoader

from otx.mpa.modules.datasets.samplers.balanced_sampler import BalancedSampler
from otx.mpa.modules.datasets.samplers.cls_incr_sampler import ClsIncrSampler
from otx.mpa.utils.logger import get_logger

logger = get_logger()


@HOOKS.register_module()
class TaskAdaptHook(Hook):
    """Task Adaptation Hook for Task-Inc & Class-Inc

    Args:
        src_classes (list): A list of old classes used in the existing model
        dst_classes (list): A list of classes including new_classes to be newly learned
        model_type (str): Types of models used for learning
        sampler_flag (bool): Flag about using ClsIncrSampler
        efficient_mode (bool): Flag about using efficient mode sampler
    """

    def __init__(
        self,
        src_classes,
        dst_classes,
        model_type="FasterRCNN",
        sampler_flag=False,
        sampler_type="cls_incr",
        efficient_mode=False,
    ):
        self.src_classes = src_classes
        self.dst_classes = dst_classes
        self.model_type = model_type
        self.sampler_flag = sampler_flag
        self.sampler_type = sampler_type
        self.efficient_mode = efficient_mode

        logger.info(f"Task Adaptation: {self.src_classes} => {self.dst_classes}")
        logger.info(f"- Efficient Mode: {self.efficient_mode}")
        logger.info(f"- Sampler type: {self.sampler_type}")
        logger.info(f"- Sampler flag: {self.sampler_flag}")

    def before_epoch(self, runner):
        if self.sampler_flag:
            dataset = runner.data_loader.dataset
            batch_size = runner.data_loader.batch_size
            num_workers = runner.data_loader.num_workers
            collate_fn = runner.data_loader.collate_fn
            worker_init_fn = runner.data_loader.worker_init_fn
            rank, world_size = get_dist_info()
            if self.sampler_type == "balanced":
                sampler = BalancedSampler(
                    dataset, batch_size, efficient_mode=self.efficient_mode, num_replicas=world_size, rank=rank
                )
            else:
                sampler = ClsIncrSampler(
                    dataset, batch_size, efficient_mode=self.efficient_mode, num_replicas=world_size, rank=rank
                )
            runner.data_loader = DataLoader(
                dataset,
                batch_size=batch_size,
                sampler=sampler,
                num_workers=num_workers,
                collate_fn=collate_fn,
                pin_memory=False,
                worker_init_fn=worker_init_fn,
            )



"""This module implements the Dataset entity."""
# Copyright (C) 2021-2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#

# pylint: disable=redefined-builtin, invalid-name

import collections.abc
import copy
import itertools
import logging
from enum import Enum
from typing import Iterator, List, Optional, Union, cast, overload

from bson.objectid import ObjectId

from otx.api.entities.annotation import AnnotationSceneEntity, AnnotationSceneKind
from otx.api.entities.dataset_item import DatasetItemEntity
from otx.api.entities.id import ID
from otx.api.entities.label import LabelEntity
from otx.api.entities.subset import Subset

logger = logging.getLogger(__name__)


class DatasetPurpose(Enum):
    """Describes the purpose for the dataset.

    This makes it possible to identify datasets for a particular use.
    """

    #: used for user inference.
    # Prediction will generate AnnotationSceneKind.PREDICTION
    INFERENCE = 0
    #: used for training.
    # AnnotationScene contains user annotation of type AnnotationSceneKind.ANNOTATION
    TRAINING = 1
    #: used for pre-evaluation, evaluation, or testing.
    # Prediction will generate AnnotationSceneKind.EVALUATION
    EVALUATION = 2
    #: used for generating output stage.
    # Prediction will generate AnnotationSceneKind.PREDICTION
    GENERATING_OUTPUT = 3
    #: used for dataset slices which are used for analysis.
    # Prediction will generate AnnotationSceneKind.INTERMEDIATE
    TEMPORARY_DATASET = 4
    #: used for task analysis.
    # Prediction will generate AnnotationSceneKind.TASK_PREDICTION
    TASK_INFERENCE = 5

    def __str__(self):
        """Returns the dataset purpose as string."""
        return str(self.name)


class DatasetIterator(collections.abc.Iterator):
    """This DatasetIterator iterates over the dataset lazily.

    Implements collections.abc.Iterator.

    Args:
        dataset (DatasetEntity): Dataset to iterate over.
    """

    def __init__(self, dataset: "DatasetEntity"):
        self.dataset = dataset
        self.index = 0

    def __next__(self) -> DatasetItemEntity:
        """Returns the next dataset item.

        Raises:
            StopIteration: if the end of the dataset is reached.

        Returns:
            DatasetItemEntity: Dataset item.
        """
        if self.index >= len(self.dataset):
            raise StopIteration
        item = self.dataset[self.index]
        self.index += 1
        return item


class DatasetEntity:
    """A dataset consists of a list of DatasetItemEntities and a purpose.

    ## With dataset items

        This way assumes the dataset item entities are constructed before the dataset entity is made.

        >>> from otx.api.entities.image import Image
        >>> from otx.api.entities.annotation import NullAnnotationSceneEntity
        >>> from otx.api.entities.dataset_item import DatasetItemEntity
        >>> item = DatasetItemEntity(media=Image(file_path="image.jpg"), annotation_scene=NullAnnotationSceneEntity())
        >>> dataset = DatasetEntity(items=[item])

    ## Iterate over dataset

        Regardless of the instantiation method chosen, the Dataset will work the same.
        The dataset can be iterated:

        >>> dataset = DatasetEntity(items=[item_1])
        >>> for dataset_item in dataset:
        ...     print(dataset_item)
        DatasetItemEntity(
            media=Image(image.jpg, width=640, height=480),
            annotation_scene=NullAnnotationSceneEntity(),
            roi=Annotation(
                shape=Rectangle(
                    x=0.0,
                    y=0.0,
                    width=1.0,
                    height=1.0
                    ),
                    labels=[],
                    id=6149e454893b7ebbe3a8faf6
                ),
            subset=NONE
        )

        A particular item can also be fetched:

        >>> first_item = dataset[0]

        Or a slice:

        >>> first_ten = dataset[:10]
        >>> last_ten = dataset[-10:]

    ## Get a subset of Dataset

        To get the test data for validating the network:

        >>> dataset = DatasetEntity()
        >>> testing_subset = dataset.get_subset(Subset.TESTING)

        This subset is also a DatasetEntity. The entities in the subset dataset refer to the same entities as
        in the original dataset. Altering one of the objects in the subset, will also alter them in the original.

    Args:
        items (Optional[List[DatasetItemEntity]]): A list of dataset items to create dataset with. Defaults to None.
        purpose (DatasetPurpose): Purpose for dataset. Refer to :class:`DatasetPurpose` for more info.
            Defaults to DatasetPurpose.INFERENCE.
    """

    def __init__(
        self,
        items: Optional[List[DatasetItemEntity]] = None,
        purpose: DatasetPurpose = DatasetPurpose.INFERENCE,
    ):
        self._items = [] if items is None else items
        self._purpose = purpose

    @property
    def purpose(self) -> DatasetPurpose:
        """Returns the DatasetPurpose. For example DatasetPurpose.ANALYSIS.

        Returns:
            DatasetPurpose
        """
        return self._purpose

    @purpose.setter
    def purpose(self, value: DatasetPurpose) -> None:
        self._purpose = value

    def _fetch(self, key: Union[slice, int]) -> Union[DatasetItemEntity, List[DatasetItemEntity]]:
        """Fetch the given entity/entities from the items.

        Helper function for __getitem__

        Args:
            key (Union[slice, int]): Key called on the dataset. E.g. int (dataset[0]) or slice (dataset[5:9])

        Raises:
            TypeError: If key is not a slice or int.

        Returns:
            Union[DatasetItemEntity, List[DatasetItemEntity]]: The entity/entities requested.
        """
        if isinstance(key, list):
            return [self._fetch(ii) for ii in key]  # type: ignore
        if isinstance(key, slice):
            # Get the start, stop, and step from the slice
            return [self._fetch(ii) for ii in range(*key.indices(len(self._items)))]  # type: ignore
        if isinstance(key, int):
            return self._items[key]
        raise TypeError(
            f"Instance of type `{type(key).__name__}` cannot be used to access Dataset items. "
            f"Only slice and int are supported"
        )

    def __repr__(self):
        """Returns string representation of the dataset."""
        return f"{self.__class__.__name__}(items={self._items}, purpose={self.purpose})"

    def __str__(self):
        """Returns string representation of the dataset."""
        return f"{self.__class__.__name__}(size={len(self)}, purpose={self.purpose})"

    def __len__(self):
        """Returns the number of items in the dataset."""
        return len(self._items)

    def __eq__(self, other: object) -> bool:
        """Checks whether the dataset is equal to the operand.

        Args:
            other (DatasetEntity): the dataset operand to the equal operator.

        Returns:
            bool: True if the datasets are equal
        """
        if isinstance(other, DatasetEntity) and len(self) == len(other) and self.purpose == other.purpose:
            return all(self[i] == other[i] for i in range(len(self)))
        return False

    def __add__(self, other: Union["DatasetEntity", List[DatasetItemEntity]]) -> "DatasetEntity":
        """Returns a new dataset which contains the items of self added with the input dataset.

        Note that additional info of the dataset might be incoherent to the addition operands.

        Args:
            other (Union[DatasetEntity, List[DatasetItemEntity]]): dataset to be added to output

        Returns:
            DatasetEntity: new dataset with the items of self added with the input dataset
        """
        items: List[DatasetItemEntity]

        if isinstance(other, DatasetEntity):
            items = self._items + list(other)
        elif isinstance(other, list):
            items = self._items + [o for o in other if isinstance(o, DatasetItemEntity)]
        else:
            raise ValueError(f"Cannot add other of type {type(other)}")

        return DatasetEntity(items=items, purpose=self.purpose)

    @overload
    def __getitem__(self, key: int) -> DatasetItemEntity:
        """Returns the item at the given index."""
        return cast(DatasetItemEntity, self._fetch(key))

    @overload  # Overload for proper type hinting of indexing on dataset.
    def __getitem__(self, key: slice) -> List[DatasetItemEntity]:
        """Returns items for the given slice."""
        return cast(List[DatasetItemEntity], self._fetch(key))

    def __getitem__(self, key: Union[slice, int]) -> Union["DatasetItemEntity", List["DatasetItemEntity"]]:
        """Return a DatasetItemEntity or a list of DatasetItemEntity, given a slice or an integer.

        Example:
            Given an integer index:

            >>> dataset = DatasetEntity(items=[...])
            >>> first_item = dataset[0]

            Or a slice:

            >>> first_ten = dataset[0:9]
            >>> last_ten = dataset[-9:]

        Args:
            key (Union[slice, int]): key to fetch. Should be `slice` or `int`

        Returns:
            Union["DatasetItemEntity", List["DatasetItemEntity"]]: List of DatasetItemEntity or single DatasetItemEntity
        """
        return self._fetch(key)

    def __iter__(self) -> Iterator[DatasetItemEntity]:
        """Return an iterator for the DatasetEntity.

        This iterator is able to iterate over the DatasetEntity lazily.

        Returns:
            DatasetIterator: DatasetIterator instance
        """
        return DatasetIterator(self)

    def with_empty_annotations(
        self, annotation_kind: AnnotationSceneKind = AnnotationSceneKind.PREDICTION
    ) -> "DatasetEntity":
        """Produces a new dataset with empty annotation objects (no shapes or labels).

        This is a convenience function to generate a dataset with empty annotations from another dataset.
        This is particularly useful for evaluation on validation data and to build resultsets.

        Assume a dataset containing user annotations.

        >>> labeled_dataset = Dataset()  # user annotated dataset

        Then, we want to see the performance of our task on this labeled_dataset,
        which means we need to create a new dataset to be passed for analysis.

        >>> prediction_dataset = labeled_dataset.with_empty_annotations()

        Later, we can pass this prediction_dataset to the task analysis function.
        By pairing the labeled_dataset and the prediction_dataset, the resultset can then be constructed.
        Refer to `otx.api.entities.resultset.ResultSetEntity` for more info.

        Args:
            annotation_kind (AnnotationSceneKind): Sets the empty annotation to this kind.
                Defaults to AnnotationSceneKind.PREDICTION

        Returns:
            DatasetEntity: a new dataset containing the same items, with empty annotation objects.
        """
        new_dataset = DatasetEntity(purpose=self.purpose)
        for dataset_item in self:
            if isinstance(dataset_item, DatasetItemEntity):
                empty_annotation = AnnotationSceneEntity(annotations=[], kind=annotation_kind)

                # reset ROI
                roi = copy.copy(dataset_item.roi)
                roi.id_ = ID(ObjectId())
                roi.set_labels([])

                new_dataset_item = DatasetItemEntity(
                    media=dataset_item.media,
                    annotation_scene=empty_annotation,
                    roi=roi,
                    subset=dataset_item.subset,
                    metadata=dataset_item.get_metadata(),
                )
                new_dataset.append(new_dataset_item)
        return new_dataset

    def get_subset(self, subset: Subset) -> "DatasetEntity":
        """Returns a new DatasetEntity with just the dataset items matching the subset.

        This subset is also a DatasetEntity. The dataset items in the subset dataset are the same dataset items as
        in the original dataset.
        Altering one of the objects in the output of this function, will also alter them in the original.

        Example:
            >>> dataset = DatasetEntity()
            >>> training_subset = dataset.get_subset(Subset.TRAINING)

        Args:
            subset (Subset): `Subset` to return.

        Returns:
            DatasetEntity: DatasetEntity with items matching subset
        """
        dataset = DatasetEntity(
            items=[item for item in self._items if item.subset == subset],
            purpose=self.purpose,
        )
        return dataset

    def remove(self, item: DatasetItemEntity) -> None:
        """Remove an item from the items.

        This function calls remove_at_indices function.

        Args:
            item (DatasetItemEntity): the item to be deleted.

        Raises:
            ValueError: if the input item is not in the dataset
        """
        index = self._items.index(item)
        self.remove_at_indices([index])

    def append(self, item: DatasetItemEntity) -> None:
        """Append a DatasetItemEntity to the dataset.

        Example:
            Appending a dataset item to a dataset

            >>> from otx.api.entities.image import Image
            >>> from otx.api.entities.annotation import NullAnnotationSceneEntity
            >>> from otx.api.entities.dataset_item import DatasetItemEntity
            >>> dataset = DatasetEntity()
            >>> media = Image(file_path='image.jpg')
            >>> annotation = NullAnnotationSceneEntity()
            >>> dataset_item = DatasetItemEntity(media=media, annotation_scene=annotation)
            >>> dataset.append(dataset_item)

        Args:
            item (DatasetItemEntity): item to append
        """

        if item.media is None:
            raise ValueError("Media in dataset item cannot be None")
        self._items.append(item)

    def sort_items(self) -> None:
        """Order the dataset items. Does nothing here, but may be overridden in child classes.

        Returns:
            None
        """

    def remove_at_indices(self, indices: List[int]) -> None:
        """Delete items based on the `indices`.

        Args:
            indices (List[int]): the indices of the items that will be deleted from the items.
        """
        indices.sort(reverse=True)  # sort in descending order
        for i_item in indices:
            del self._items[i_item]

    def get_labels(self, include_empty: bool = False) -> List[LabelEntity]:
        """Returns the list of all unique labels that are in the dataset.

        Note: This does not respect the ROI of the dataset items.

        Args:
            include_empty (bool): set to True to include empty label (if exists) in the output. Defaults to False.

        Returns:
            List[LabelEntity]: list of labels that appear in the dataset
        """
        label_set = set(itertools.chain(*[item.annotation_scene.get_labels(include_empty) for item in self]))
        return list(label_set)



# Copyright (C) 2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#

from mmcv import ConfigDict

from otx.mpa.det.stage import DetectionStage
from otx.mpa.utils.config_utils import update_or_add_custom_hook
from otx.mpa.utils.logger import get_logger

logger = get_logger()


class IncrDetectionStage(DetectionStage):
    """Patch config to support incremental learning for object detection"""

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    def configure_task(self, cfg, training, **kwargs):
        """Patch config to support incremental learning"""
        super().configure_task(cfg, training, **kwargs)
        if "task_adapt" in cfg and self.task_adapt_type == "mpa":
            self.configure_task_adapt_hook(cfg)

    def configure_task_adapt_hook(self, cfg):
        """Add TaskAdaptHook for sampler."""
        sampler_flag = True
        if len(set(self.org_model_classes) & set(self.model_classes)) == 0 or set(self.org_model_classes) == set(
            self.model_classes
        ):
            sampler_flag = False
        update_or_add_custom_hook(
            cfg,
            ConfigDict(
                type="TaskAdaptHook",
                src_classes=self.org_model_classes,
                dst_classes=self.model_classes,
                model_type=cfg.model.type,
                sampler_flag=sampler_flag,
                efficient_mode=cfg["task_adapt"].get("efficient_mode", False),
            ),
        )



# Copyright (C) 2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#

# code in this file is adpated from
# https://github.com/ildoonet/pytorch-randaugment/blob/master/RandAugment/augmentations.py
# https://github.com/google-research/fixmatch/blob/master/third_party/auto_augment/augmentations.py
# https://github.com/google-research/fixmatch/blob/master/libml/ctaugment.py
import random

import numpy as np
import PIL
from mmcls.datasets.builder import PIPELINES

PARAMETER_MAX = 10


def AutoContrast(img, **kwarg):
    return PIL.ImageOps.autocontrast(img), None


def Brightness(img, v, max_v, bias=0):
    v = _float_parameter(v, max_v) + bias
    return PIL.ImageEnhance.Brightness(img).enhance(v), v


def Color(img, v, max_v, bias=0):
    v = _float_parameter(v, max_v) + bias
    return PIL.ImageEnhance.Color(img).enhance(v), v


def Contrast(img, v, max_v, bias=0):
    v = _float_parameter(v, max_v) + bias
    return PIL.ImageEnhance.Contrast(img).enhance(v), v


def Cutout(img, v, max_v, bias=0):
    if v == 0:
        return img
    v = _float_parameter(v, max_v) + bias
    v = int(v * min(img.size))
    return CutoutAbs(img, v), v


def CutoutAbs(img, v, **kwarg):
    w, h = img.size
    x0 = np.random.uniform(0, w)
    y0 = np.random.uniform(0, h)
    x0 = int(max(0, x0 - v / 2.0))
    y0 = int(max(0, y0 - v / 2.0))
    x1 = int(min(w, x0 + v))
    y1 = int(min(h, y0 + v))
    xy = (x0, y0, x1, y1)
    # gray
    color = (127, 127, 127)
    img = img.copy()
    PIL.ImageDraw.Draw(img).rectangle(xy, color)
    return img, xy, color


def Equalize(img, **kwarg):
    return PIL.ImageOps.equalize(img), None


def Identity(img, **kwarg):
    return img, None


def Posterize(img, v, max_v, bias=0):
    v = _int_parameter(v, max_v) + bias
    return PIL.ImageOps.posterize(img, v), v


def Rotate(img, v, max_v, bias=0):
    v = _int_parameter(v, max_v) + bias
    if random.random() < 0.5:
        v = -v
    return img.rotate(v), v


def Sharpness(img, v, max_v, bias=0):
    v = _float_parameter(v, max_v) + bias
    return PIL.ImageEnhance.Sharpness(img).enhance(v), v


def ShearX(img, v, max_v, bias=0):
    v = _float_parameter(v, max_v) + bias
    if random.random() < 0.5:
        v = -v
    return img.transform(img.size, PIL.Image.AFFINE, (1, v, 0, 0, 1, 0)), v


def ShearY(img, v, max_v, bias=0):
    v = _float_parameter(v, max_v) + bias
    if random.random() < 0.5:
        v = -v
    return img.transform(img.size, PIL.Image.AFFINE, (1, 0, 0, v, 1, 0)), v


def Solarize(img, v, max_v, bias=0):
    v = _int_parameter(v, max_v) + bias
    return PIL.ImageOps.solarize(img, 256 - v), v


def TranslateX(img, v, max_v, bias=0):
    v = _float_parameter(v, max_v) + bias
    if random.random() < 0.5:
        v = -v
    v = int(v * img.size[0])
    return img.transform(img.size, PIL.Image.AFFINE, (1, 0, v, 0, 1, 0)), v


def TranslateY(img, v, max_v, bias=0):
    v = _float_parameter(v, max_v) + bias
    if random.random() < 0.5:
        v = -v
    v = int(v * img.size[1])
    return img.transform(img.size, PIL.Image.AFFINE, (1, 0, 0, 0, 1, v)), v


def _float_parameter(v, max_v):
    return float(v) * max_v / PARAMETER_MAX


def _int_parameter(v, max_v):
    return int(v * max_v / PARAMETER_MAX)


rand_augment_pool = [
    (AutoContrast, None, None),
    (Brightness, 0.9, 0.05),
    (Color, 0.9, 0.05),
    (Contrast, 0.9, 0.05),
    (Equalize, None, None),
    (Identity, None, None),
    (Posterize, 4, 4),
    (Rotate, 30, 0),
    (Sharpness, 0.9, 0.05),
    (ShearX, 0.3, 0),
    (ShearY, 0.3, 0),
    (Solarize, 256, 0),
    (TranslateX, 0.3, 0),
    (TranslateY, 0.3, 0),
]

# TODO: [Jihwan]: Can be removed by mmcls.datasets.pipeline.auto_augment Line 95 RandAugment class
@PIPELINES.register_module()
class MPARandAugment(object):
    def __init__(self, n, m, cutout=16):
        assert n >= 1
        assert 1 <= m <= 10
        self.n = n
        self.m = m
        self.cutout = cutout
        self.augment_pool = rand_augment_pool

    def __call__(self, results):
        for key in results.get("img_fields", ["img"]):
            img = results[key]
            if not PIL.Image.isImageType(img):
                img = PIL.Image.fromarray(results[key])
            ops = random.choices(self.augment_pool, k=self.n)
            for op, max_v, bias in ops:
                v = np.random.randint(1, self.m)
                if random.random() < 0.5:
                    img, v = op(img, v=v, max_v=max_v, bias=bias)
                    results["rand_mc_{}".format(op.__name__)] = v
            img, xy, color = CutoutAbs(img, self.cutout)
            results["CutoutAbs"] = (xy, self.cutout, color)
            results[key] = np.array(img)
        return results



# Copyright (C) 2021-2023 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#

import pytest

from otx.algorithms.detection.tasks import DetectionNNCFTask
from otx.api.configuration.configurable_parameters import ConfigurableParameters
from otx.api.entities.datasets import DatasetEntity
from otx.api.entities.label_schema import LabelSchemaEntity
from otx.api.entities.model import ModelConfiguration, ModelEntity
from otx.api.usecases.tasks.interfaces.export_interface import ExportType
from otx.api.usecases.tasks.interfaces.optimization_interface import OptimizationType
from tests.test_suite.e2e_test_system import e2e_pytest_unit
from tests.unit.api.parameters_validation.validation_helper import (
    check_value_error_exception_raised,
)


class MockNNCFTask(DetectionNNCFTask):
    def __init__(self):
        pass


class TestNNCFTaskInputParamsValidation:
    @staticmethod
    def model():
        model_configuration = ModelConfiguration(
            configurable_parameters=ConfigurableParameters(header="header", description="description"),
            label_schema=LabelSchemaEntity(),
        )
        return ModelEntity(train_dataset=DatasetEntity(), configuration=model_configuration)

    @e2e_pytest_unit
    def test_nncf_detection_task_init_params_validation(self):
        """
        <b>Description:</b>
        Check DetectionNNCFTask object initialization parameters validation

        <b>Input data:</b>
        DetectionNNCFTask object initialization parameters with unexpected type

        <b>Expected results:</b>
        Test passes if ValueError exception is raised when unexpected type object is specified as
        DetectionNNCFTask object initialization parameter
        """
        with pytest.raises(ValueError):
            DetectionNNCFTask(task_environment="unexpected string")  # type: ignore

    @e2e_pytest_unit
    def test_nncf_detection_task_optimize_params_validation(self):
        """
        <b>Description:</b>
        Check DetectionNNCFTask object "optimize" method input parameters validation

        <b>Input data:</b>
        DetectionNNCFTask object. "optimize" method unexpected-type input parameters

        <b>Expected results:</b>
        Test passes if ValueError exception is raised when unexpected type object is specified as
        input parameter for "optimize" method
        """
        task = MockNNCFTask()
        correct_values_dict = {
            "optimization_type": OptimizationType.NNCF,
            "dataset": DatasetEntity(),
            "output_model": self.model(),
        }
        unexpected_str = "unexpected string"
        unexpected_values = [
            # Unexpected string is specified as "optimization_type" parameter
            ("optimization_type", unexpected_str),
            # Unexpected string is specified as "dataset" parameter
            ("dataset", unexpected_str),
            # Unexpected string is specified as "output_model" parameter
            ("output_model", unexpected_str),
            # Unexpected string is specified as "optimization_parameters" parameter
            ("optimization_parameters", unexpected_str),
        ]
        check_value_error_exception_raised(
            correct_parameters=correct_values_dict,
            unexpected_values=unexpected_values,
            class_or_function=task.optimize,
        )

    @e2e_pytest_unit
    def test_nncf_detection_task_export_params_validation(self):
        """
        <b>Description:</b>
        Check DetectionNNCFTask object "export" method input parameters validation

        <b>Input data:</b>
        DetectionNNCFTask object. "export" method unexpected-type input parameters

        <b>Expected results:</b>
        Test passes if ValueError exception is raised when unexpected type object is specified as
        input parameter for "export" method
        """
        task = MockNNCFTask()
        correct_values_dict = {
            "export_type": ExportType.OPENVINO,
            "output_model": self.model(),
        }
        unexpected_str = "unexpected string"
        unexpected_values = [
            # Unexpected string is specified as "export_type" parameter
            ("export_type", unexpected_str),
            # Unexpected string is specified as "output_model" parameter
            ("output_model", unexpected_str),
        ]
        check_value_error_exception_raised(
            correct_parameters=correct_values_dict,
            unexpected_values=unexpected_values,
            class_or_function=task.export,
        )

    @e2e_pytest_unit
    def test_nncf_detection_task_save_model_params_validation(self):
        """
        <b>Description:</b>
        Check DetectionNNCFTask object "save_model" method input parameters validation

        <b>Input data:</b>
        DetectionNNCFTask object, "output_model" non-ModelEntity object

        <b>Expected results:</b>
        Test passes if ValueError exception is raised when unexpected type object is specified as
        input parameter for "save_model" method
        """
        task = MockNNCFTask()
        with pytest.raises(ValueError):
            task.save_model(output_model="unexpected string")  # type: ignore



# Copyright (C) 2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#

from mmcv.runner import HOOKS, EpochBasedRunner, Hook

from otx.mpa.utils.logger import get_logger

logger = get_logger()


@HOOKS.register_module()
class CancelInterfaceHook(Hook):
    def __init__(self, init_callback: callable, interval=5):
        self.on_init_callback = init_callback
        self.runner = None
        self.interval = interval

    def cancel(self):
        logger.info("CancelInterfaceHook.cancel() is called.")
        if self.runner is None:
            logger.warning("runner is not configured yet. ignored this request.")
            return

        if self.runner.should_stop:
            logger.warning("cancel already requested.")
            return

        if isinstance(self.runner, EpochBasedRunner):
            epoch = self.runner.epoch
            self.runner._max_epochs = epoch  # Force runner to stop by pretending it has reached it's max_epoch
        self.runner.should_stop = True  # Set this flag to true to stop the current training epoch
        logger.info("requested stopping to the runner")

    def before_run(self, runner):
        self.runner = runner
        self.on_init_callback(self)

    def after_run(self, runner):
        self.runner = None



"""NNCF Task of OTX Segmentation."""

# Copyright (C) 2022 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions
# and limitations under the License.

from functools import partial
from typing import List, Optional

import otx.algorithms.segmentation.adapters.mmseg.nncf.patches  # noqa: F401  # pylint: disable=unused-import
from otx.algorithms.common.tasks.nncf_base import NNCFBaseTask
from otx.algorithms.segmentation.adapters.mmseg.nncf import build_nncf_segmentor
from otx.api.entities.datasets import DatasetEntity
from otx.api.entities.metrics import (
    CurveMetric,
    InfoMetric,
    LineChartInfo,
    MetricsGroup,
    Performance,
    ScoreMetric,
    VisualizationInfo,
    VisualizationType,
)
from otx.api.entities.model import ModelEntity
from otx.api.entities.optimization_parameters import OptimizationParameters
from otx.mpa.utils.logger import get_logger

from .inference import SegmentationInferenceTask

logger = get_logger()


class SegmentationNNCFTask(NNCFBaseTask, SegmentationInferenceTask):  # pylint: disable=too-many-ancestors
    """SegmentationNNCFTask."""

    def _initialize_post_hook(self, options=None):
        super()._initialize_post_hook(options)

        export = options.get("export", False)
        options["model_builder"] = partial(
            self.model_builder,
            nncf_model_builder=build_nncf_segmentor,
            return_compression_ctrl=False,
            is_export=export,
        )

    def _optimize(
        self,
        dataset: DatasetEntity,
        optimization_parameters: Optional[OptimizationParameters] = None,
    ):
        results = self._run_task(
            "SegTrainer",
            mode="train",
            dataset=dataset,
            parameters=optimization_parameters,
        )
        return results

    def _optimize_post_hook(
        self,
        dataset: DatasetEntity,
        output_model: ModelEntity,
    ):
        # Get training metrics group from learning curves
        training_metrics, best_score = self._generate_training_metrics_group(self._learning_curves)
        performance = Performance(
            score=ScoreMetric(value=best_score, name=self.metric),
            dashboard_metrics=training_metrics,
        )

        logger.info(f"Final model performance: {str(performance)}")
        output_model.performance = performance

    def _generate_training_metrics_group(self, learning_curves):
        """Get Training metrics (epochs & scores).

        Parses the mmsegmentation logs to get metrics from the latest training run
        :return output List[MetricsGroup]
        """
        output: List[MetricsGroup] = []
        # Model architecture
        architecture = InfoMetric(name="Model architecture", value=self._model_name)
        visualization_info_architecture = VisualizationInfo(
            name="Model architecture", visualisation_type=VisualizationType.TEXT
        )
        output.append(
            MetricsGroup(
                metrics=[architecture],
                visualization_info=visualization_info_architecture,
            )
        )
        # Learning curves
        best_score = -1
        for key, curve in learning_curves.items():
            metric_curve = CurveMetric(xs=curve.x, ys=curve.y, name=key)
            if key == f"val/{self.metric}":
                best_score = max(curve.y)
            visualization_info = LineChartInfo(name=key, x_axis_label="Epoch", y_axis_label=key)
            output.append(MetricsGroup(metrics=[metric_curve], visualization_info=visualization_info))
        return output, best_score



"""Data configuration for AVA dataset."""

# Copyright (C) 2022 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions
# and limitations under the License.

# pylint: disable=invalid-name


_base_ = ["./data_pipeline.py"]

# FIXME Only changes from base is frame interval of SampleAVAFrames
img_norm_cfg = dict(mean=[123.675, 116.28, 103.53], std=[58.395, 57.12, 57.375], to_bgr=False)

train_pipeline = [
    dict(type="SampleAVAFrames", clip_len=32, frame_interval=2),
    dict(type="RawFrameDecode"),
    dict(type="RandomRescale", scale_range=(256, 320)),
    dict(type="RandomCrop", size=256),
    dict(type="Flip", flip_ratio=0.5),
    dict(type="Normalize", **img_norm_cfg),
    dict(type="FormatShape", input_format="NCTHW", collapse=True),
    # Rename is needed to use mmdet detectors
    dict(type="Rename", mapping=dict(imgs="img")),
    dict(type="ToTensor", keys=["img", "proposals", "gt_bboxes", "gt_labels"]),
    dict(type="ToDataContainer", fields=[dict(key=["proposals", "gt_bboxes", "gt_labels"], stack=False)]),
    dict(type="Collect", keys=["img", "proposals", "gt_bboxes", "gt_labels"], meta_keys=["scores", "entity_ids"]),
]
# The testing is w/o. any cropping / flipping
val_pipeline = [
    dict(type="SampleAVAFrames", clip_len=32, frame_interval=2, test_mode=True),
    dict(type="RawFrameDecode"),
    dict(type="Resize", scale=(-1, 256)),
    dict(type="Normalize", **img_norm_cfg),
    dict(type="FormatShape", input_format="NCTHW", collapse=True),
    # Rename is needed to use mmdet detectors
    dict(type="Rename", mapping=dict(imgs="img")),
    dict(type="ToTensor", keys=["img", "proposals"]),
    dict(type="ToDataContainer", fields=[dict(key="proposals", stack=False)]),
    dict(type="Collect", keys=["img", "proposals"], meta_keys=["scores", "img_shape"], nested=True),
]

data = dict(
    val_dataloader=dict(videos_per_gpu=1),
    test_dataloader=dict(videos_per_gpu=1),
    train=dict(
        pipeline=train_pipeline,
        person_det_score_thr=0.9,
        fps=30,
    ),
    val=dict(
        pipeline=val_pipeline,
        person_det_score_thr=0.9,
        fps=30,
        test_mode=True,
    ),
)
data["test"] = data["val"]



# Copyright (C) 2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#

from dataclasses import dataclass, field

from .builder import OPS
from .op import Attribute, Operation


@dataclass
class TopKV3Attribute(Attribute):
    axis: int
    mode: str
    sort: str
    index_element_type: str = field(default="i32")

    def __post_init__(self):
        super().__post_init__()
        valid_mode = ["min", "max"]
        if self.mode not in valid_mode:
            raise ValueError(f"Invalid mode {self.mode}. " f"It must be one of {valid_mode}.")

        valid_sort = ["value", "index", "none"]
        if self.sort not in valid_sort:
            raise ValueError(f"Invalid sort {self.sort}. " f"It must be one of {valid_sort}.")

        valid_index_element_type = ["i32", "i64"]
        if self.index_element_type not in valid_index_element_type:
            raise ValueError(
                f"Invalid index_element_type {self.index_element_type}. "
                f"It must be one of {valid_index_element_type}."
            )


@OPS.register()
class TopKV3(Operation[TopKV3Attribute]):
    TYPE = "TopK"
    VERSION = 3
    ATTRIBUTE_FACTORY = TopKV3Attribute

    def forward(self, input, k):
        raise NotImplementedError
        #  values, indices = torch.topk(
        #      input=input,
        #      k=k,
        #      dim=self.attrs.axis,
        #      largest=self.attrs.mode == "max",
        #      sorted=True,
        #  )
        #
        #  if self.attrs.sort == "index":
        #      sorted = torch.argsort(indices)
        #      indices = indices[sorted]
        #      values = values[sorted]
        #
        #  return values, indices


@dataclass
class NonMaxSuppressionV5Attribute(Attribute):
    box_encoding: str = field(default="corner")
    sort_result_descending: bool = field(default=True)
    output_type: str = field(default="i64")


@OPS.register()
class NonMaxSuppressionV5(Operation[NonMaxSuppressionV5Attribute]):
    TYPE = "NonMaxSuppression"
    VERSION = 5
    ATTRIBUTE_FACTORY = NonMaxSuppressionV5Attribute

    def forward(
        self,
        boxes,
        scores,
        max_output_boxes_per_class,
        iou_threshold=0,
        score_threshold=0,
        soft_nms_sigma=0,
    ):
        raise NotImplementedError


@dataclass
class NonMaxSuppressionV9Attribute(Attribute):
    box_encoding: str = field(default="corner")
    sort_result_descending: bool = field(default=True)
    output_type: str = field(default="i64")


@OPS.register()
class NonMaxSuppressionV9(Operation[NonMaxSuppressionV9Attribute]):
    TYPE = "NonMaxSuppression"
    VERSION = 9
    ATTRIBUTE_FACTORY = NonMaxSuppressionV9Attribute

    def forward(
        self,
        boxes,
        scores,
        max_output_boxes_per_class,
        iou_threshold=0,
        score_threshold=0,
        soft_nms_sigma=0,
    ):
        raise NotImplementedError



# Copyright (C) 2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#

import math
import os
import tempfile
from collections import OrderedDict
from copy import deepcopy
from typing import Callable, List, Optional, Union

import openvino.runtime as ov
import torch
from torch.nn import init

from otx.mpa.utils.logger import get_logger

from ..graph import Graph
from ..graph.utils import (
    handle_merging_into_batchnorm,
    handle_paired_batchnorm,
    handle_reshape,
)
from ..ops import OPS
from ..utils import load_ov_model, normalize_name

logger = get_logger()


CONNECTION_SEPARATOR = "||"


class OVModel(torch.nn.Module):
    def __init__(
        self,
        model_path_or_model: Union[str, ov.Model] = None,
        weight_path: Optional[str] = None,
        inputs: Union[str, List[str]] = [],
        outputs: Union[str, List[str]] = [],
        features_to_keep: Optional[List] = None,
        remove_normalize: bool = False,
        merge_bn: bool = True,
        paired_bn: bool = True,
        init_weight: Union[bool, Callable] = False,
        verify_shape: bool = True,
    ):
        super().__init__()
        self._model_path_or_model = model_path_or_model
        self._weight_path = weight_path
        self._remove_normalize = remove_normalize
        self._features_to_keep = features_to_keep
        self._merge_bn = merge_bn
        self._paired_bn = paired_bn
        self._init_weight = init_weight
        self._verify_shape = verify_shape

        self._inputs = []
        self._outputs = []
        self._feature_dict = OrderedDict()

        # build graph
        graph = self.build_graph(model_path_or_model, weight_path)
        self._graph = graph
        if remove_normalize:
            graph.remove_normalize_nodes()

        # handle inputs
        if inputs:
            inputs = inputs if isinstance(inputs, list) else [inputs]
            assert all(isinstance(i, str) for i in inputs), f"input must be string but {inputs} is given"
            inputs = self.build_custom_inputs(graph, deepcopy(inputs))
        else:
            inputs = [node.name for node in graph.get_nodes_by_types(["Parameter"])]
        self._inputs = inputs

        # handle outputs
        if outputs:
            outputs = outputs if isinstance(outputs, list) else [outputs]
            assert all(isinstance(i, str) for i in outputs), f"input must be string but {outputs} is given"
            outputs = self.build_custom_outputs(graph, deepcopy(outputs))
        else:
            outputs = [node.name for node in graph.get_nodes_by_types(["Result"])]
        self._outputs = outputs

        # clean up graph
        self.clean_up(graph, inputs, outputs)

        handle_reshape(graph)
        if merge_bn:
            handle_merging_into_batchnorm(graph)
        if paired_bn:
            handle_paired_batchnorm(graph, replace=True)

        # clean up graph
        self.clean_up(graph, inputs, outputs)

        # build torch module
        self.model = self.build_torch_module(graph)

        if init_weight:
            if not isinstance(init_weight, Callable):

                # internal init weight
                def init_weight(m, graph):
                    from ..ops.op import Operation

                    if not isinstance(m, Operation):
                        return

                    if m.TYPE == "BatchNormInference":
                        _, gamma, beta, mean, var = list(graph.predecessors(m))
                        init.ones_(gamma.data)
                        init.zeros_(beta.data)
                        mean.data.zero_()
                        var.data.fill_(1)
                        logger.info(f"Initialize {m.TYPE} -> {m.name}")
                    elif m.TYPE in [
                        "Convolution",
                        "GroupConvolution",
                        "MatMul",
                    ]:
                        for weight in graph.predecessors(m):
                            if weight.TYPE == "Constant" and isinstance(weight.data, torch.nn.parameter.Parameter):
                                init.kaiming_uniform_(weight.data, a=math.sqrt(5))
                                logger.info(f"Initialize {m.TYPE} -> {m.name}")
                    elif m.TYPE in [
                        "Multiply",
                        "Divide",
                        "Add",
                        "Subtract",
                    ]:
                        for weight in graph.predecessors(m):
                            if weight.TYPE == "Constant" and isinstance(weight.data, torch.nn.parameter.Parameter):
                                fan_in, _ = init._calculate_fan_in_and_fan_out(weight.data)
                                bound = 1 / math.sqrt(fan_in) if fan_in > 0 else 0
                                init.uniform_(weight.data, -bound, bound)
                                logger.info(f"Initialize {m.TYPE} -> {m.name}")

            self.model.apply(lambda m: init_weight(m, graph))

        for node in self._graph.get_nodes_by_types(["Parameter"]):
            node.attrs.verify_shape = verify_shape

        input_shapes = {}
        output_shapes = {}
        for node in self._graph.get_nodes_by_types(["Parameter", "Result"]):
            if node.name in self._inputs:
                input_shapes[node.name] = node.shape[0]
            elif node.name in self._outputs:
                output_shapes[node.name] = node.shape[0]
        self._input_shapes = OrderedDict()
        self._output_shapes = OrderedDict()
        for input in self._inputs:
            self._input_shapes[input] = input_shapes[input]
        for output in self._outputs:
            self._output_shapes[output] = output_shapes[output]

    @property
    def inputs(self):
        return self._inputs

    @property
    def outputs(self):
        return self._outputs

    @property
    def features(self):
        return self._feature_dict

    @property
    def input_shapes(self):
        return self._input_shapes

    @property
    def output_shapes(self):
        return self._output_shapes

    @staticmethod
    def build_graph(model_path_or_model, weight_path=None):
        with tempfile.TemporaryDirectory() as tempdir:
            if isinstance(model_path_or_model, ov.Model):
                assert weight_path is None, "if openvino model is given 'weight_path' must be None"
                ov.serialize(
                    model_path_or_model,
                    os.path.join(tempdir, "model.xml"),
                    os.path.join(tempdir, "model.bin"),
                )
                model_path_or_model = os.path.join(tempdir, "model.xml")
                weight_path = os.path.join(tempdir, "model.bin")
            # TODO: reshape decompose ir graph
            ov_model = load_ov_model(model_path_or_model, weight_path, False)
        graph = Graph.from_ov(ov_model)
        return graph

    @staticmethod
    def build_custom_outputs(graph, outputs):
        cls_result = OPS.get_by_type_version("Result", 0)
        node_dict = OrderedDict((i.name, i) for i in graph.topological_sort())

        if not isinstance(outputs, list):
            outputs = [outputs]

        nodes_to_remove = []
        edges_to_add = {}
        for i, output in enumerate(outputs):
            output = normalize_name(output)
            output = output.split(CONNECTION_SEPARATOR)
            explicit_tgt = False

            if len(output) == 1:
                src = output[0]
                tgt = None
            elif len(output) == 2:
                src, tgt = output
                explicit_tgt = True
            else:
                raise ValueError()

            src = node_dict[src]
            if src.type == "Result":
                continue

            if explicit_tgt:
                tgt = node_dict[tgt]
            else:
                tgt = list(graph.successors(src))[0]

            output_result = f"{src.name}/result_{i}"
            outputs[i] = output_result

            if src not in edges_to_add:
                edges_to_add[src] = []

            for successor in graph.successors(src):
                if tgt == successor:
                    edges_attrs = graph.get_edge_data(src, successor)
                    assert len(edges_attrs) == 1

                    output_result = cls_result(output_result, shape=src.shape)
                    for edge_attrs in edges_attrs:
                        edges_to_add[src].append({"node_from": src, "node_to": output_result, **edge_attrs})
                if explicit_tgt and tgt != successor:
                    continue
                nodes_to_remove.append(successor)

        #  handle duplicated successors
        merge_candidates = [k for k, v in edges_to_add.items() if len(v) > 1]
        if merge_candidates:
            for merge_candidate in merge_candidates:
                edges = edges_to_add[merge_candidate]
                seen = {}
                out_ports = [edge["out_port"] for edge in edges]
                for idx in reversed(range(len(out_ports))):
                    out_port = out_ports[idx]
                    if out_port in seen:
                        edge = edges.pop(idx)
                        outputs.pop(outputs.index(edge["node_to"].name))
                    else:
                        seen[out_port] = edges[idx]

        if edges_to_add:
            for edges in edges_to_add.values():
                for edge in edges:
                    edge["in_port"] = 0
            assert set([len(edges) for edges in edges_to_add.values()]) == {1}
            edges_to_add = [edge for edges in edges_to_add.values() for edge in edges]
        else:
            edges_to_add = []

        for node in set(nodes_to_remove):
            graph.remove_node(node)
        for edge in edges_to_add:
            graph.add_edge(**edge)
        return outputs

    @staticmethod
    def build_custom_inputs(graph, inputs: Union[str, List[str]]):
        cls_param = OPS.get_by_type_version("Parameter", 0)
        node_dict = OrderedDict((i.name, i) for i in graph.topological_sort())

        if not isinstance(inputs, list):
            inputs = [inputs]

        edges_to_add = {}
        nodes_to_remove = []
        for i, input in enumerate(inputs):
            input = normalize_name(input)
            input = input.split(CONNECTION_SEPARATOR)
            explicit_src = False

            if len(input) == 1:
                src = None
                tgt = input[0]
            elif len(input) == 2:
                src, tgt = input
                explicit_src = True
            else:
                raise ValueError()

            tgt = node_dict[tgt]
            if tgt.type == "Parameter":
                continue

            if explicit_src:
                src = node_dict[src]
            else:
                src = list(graph.predecessors(tgt))[0]

            input_parameter = f"{tgt.name}/parameter_{i}"
            inputs[i] = input_parameter

            if src not in edges_to_add:
                edges_to_add[src] = []

            for predecessor in graph.predecessors(tgt):
                if src == predecessor:
                    edges_attrs = graph.get_edge_data(predecessor, tgt)
                    assert len(edges_attrs) == 1

                    # TODO: here, we force the batch dim to be dynamic
                    # it is assumed to be dim 0
                    new_shape = []
                    for shape in predecessor.shape:
                        new_shape.append([-1 if j == 0 else k for j, k in enumerate(shape)])
                    new_shape = tuple(tuple(shape) for shape in new_shape)
                    input_parameter = cls_param(input_parameter, shape=new_shape)
                    for edge_attrs in edges_attrs:
                        edges_to_add[src].append({"node_from": input_parameter, "node_to": tgt, **edge_attrs})
                if (explicit_src and src != predecessor) or predecessor.type == "Constant":
                    continue
                nodes_to_remove.append(predecessor)

        # handle duplicated predecessors
        merge_candidates = [k for k, v in edges_to_add.items() if len(v) > 1]
        if merge_candidates:
            for merge_candidate in merge_candidates:
                ctr = 0
                edges = edges_to_add[merge_candidate]
                seen = {}
                out_ports = [edge["out_port"] for edge in edges]
                for idx in reversed(range(len(out_ports))):
                    out_port = out_ports[idx]
                    if out_port in seen:
                        edge = edges.pop(idx)
                        inputs.pop(inputs.index(edge["node_from"].name))
                        edge["node_from"] = seen[out_port]["node_from"]
                        edges_to_add[f"{merge_candidate.name}_{ctr}"] = [edge]
                        ctr += 1
                    else:
                        seen[out_port] = edges[idx]

        if edges_to_add:
            for edges in edges_to_add.values():
                for edge in edges:
                    edge["out_port"] = 0
            assert set([len(edges) for edges in edges_to_add.values()]) == {1}
            edges_to_add = [edge for edges in edges_to_add.values() for edge in edges]
        else:
            edges_to_add = []

        for node in set(nodes_to_remove):
            graph.remove_node(node)
        for edge in edges_to_add:
            graph.add_edge(**edge)
        return inputs

    @staticmethod
    def clean_up(graph, inputs=[], outputs=[]):
        nodes = list(graph.topological_sort())
        nodes_to_keep = []
        for node in nodes:
            if node.name in inputs or node.name in outputs:
                nodes_to_keep.append(node)

        def get_nodes_without_successors(graph, ignores=[]):
            outputs = []
            for node in reversed(list(graph.topological_sort())):
                if not list(graph.successors(node)) and node not in ignores:
                    outputs.append(node)
            return outputs

        nodes = get_nodes_without_successors(graph, nodes_to_keep)
        while nodes:
            graph.remove_nodes_from(nodes)
            nodes = get_nodes_without_successors(graph, nodes_to_keep)

        graph.clean_up(nodes_to_keep)

    @staticmethod
    def build_torch_module(graph):
        node_dict = OrderedDict((i.name, i) for i in graph.topological_sort())
        return torch.nn.ModuleDict(list(node_dict.items()))

    def _build_forward_inputs(self, *args, **kwargs):
        inputs = {}
        if args:
            for key, arg in zip(self._inputs, args):
                inputs[key] = arg
        if kwargs:
            for key, arg in kwargs.items():
                if key in inputs:
                    raise ValueError
                inputs[key] = arg
        return inputs

    def forward(self, *args, **kwargs):
        self._feature_dict.clear()
        inputs = self._build_forward_inputs(*args, **kwargs)

        done = {}
        for node_name, node in self.model.items():
            done[node_name] = {node.name: False for node in self._graph.successors(node)}

        for node_name, node in self.model.items():
            predecessors_with_edge = list(self._graph.predecessors(node, with_edge_data=True))
            if not predecessors_with_edge:
                if node.type == "Parameter":
                    self._feature_dict[node_name] = node(inputs[node_name])
                elif node.type == "Constant":
                    self._feature_dict[node_name] = node()
                else:
                    raise ValueError(
                        f"Broken graph. Node {node_name} is a type of {node.type} " "but it has no in edges."
                    )
            else:
                input_nodes, edges = list(map(list, zip(*predecessors_with_edge)))
                input_node_names = [input_node.name for input_node in input_nodes]

                input_features = [edge["in_port"] for edges_ in edges for edge in edges_]
                assert len(input_features) == len(set(input_features))
                input_features = [None for _ in input_features]
                for idx, input_node_name in enumerate(input_node_names):
                    if self._features_to_keep is not None and input_node_name in self._features_to_keep:
                        input_feature = self._feature_dict.get(input_node_name)
                    else:
                        input_feature = self._feature_dict.pop(input_node_name)
                        done[input_node_name][node_name] = True
                        if not all(done[input_node_name].values()):
                            self._feature_dict[input_node_name] = input_feature

                    if isinstance(input_feature, tuple):
                        for edges_ in edges[idx]:
                            input_features[edges_["in_port"]] = input_feature[edges_["out_port"]]
                    else:
                        for edges_ in edges[idx]:
                            input_features[edges_["in_port"]] = input_feature
                assert all(input_feature is not None for input_feature in input_features)
                self._feature_dict[node_name] = node(*input_features)

        outputs = OrderedDict()
        for output_name in self._outputs:
            outputs[output_name] = self._feature_dict[output_name]

        return outputs



"""Utils for parsing command line arguments."""

# Copyright (C) 2021 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions
# and limitations under the License.

import argparse
import sys
from argparse import RawTextHelpFormatter
from pathlib import Path
from typing import Dict, List, Optional, Union

from otx.api.entities.model_template import ModelTemplate, parse_model_template
from otx.cli.registry import find_and_parse_model_template


def gen_param_help(hyper_parameters: Dict) -> Dict:
    """Generates help for hyper parameters section."""

    type_map = {"FLOAT": float, "INTEGER": int, "BOOLEAN": bool, "SELECTABLE": str}

    help_keys = ("header", "type", "default_value", "max_value", "min_value")

    def _gen_param_help(prefix: str, cur_params: Dict) -> Dict:
        cur_help = {}
        for k, val in cur_params.items():
            if not isinstance(val, dict):
                continue

            if "default_value" not in val.keys():
                if "visible_in_ui" in val and val["visible_in_ui"]:
                    x = _gen_param_help(prefix + f"{k}.", val)
                    cur_help.update(x)
            else:
                assert isinstance(val["default_value"], (int, float, str))
                help_str = "\n".join([f"{kk}: {val[kk]}" for kk in help_keys if kk in val.keys()])
                assert "." not in k

                cur_help.update(
                    {
                        prefix
                        + f"{k}": {
                            "default": val["default_value"],
                            "help": help_str,
                            "type": type_map[val["type"]],
                            "affects_outcome_of": val["affects_outcome_of"],
                        }
                    }
                )
        return cur_help

    return _gen_param_help("", hyper_parameters)


def gen_params_dict_from_args(
    args, override_param: Optional[List] = None, type_hint: Optional[dict] = None
) -> Dict[str, dict]:
    """Generates hyper parameters dict from parsed command line arguments."""

    params_dict: Dict[str, dict] = {}
    for param_name in dir(args):
        if not param_name.startswith("params."):
            continue
        if override_param and param_name not in override_param:
            continue

        value_type = None
        cur_dict = params_dict
        split_param_name = param_name.split(".")[1:]
        if type_hint:
            origin_key = ".".join(split_param_name)
            value_type = type_hint[origin_key].get("type", None)
        for i, k in enumerate(split_param_name):
            if k not in cur_dict:
                cur_dict[k] = {}
            if i < len(split_param_name) - 1:
                cur_dict = cur_dict[k]
            else:
                value = getattr(args, param_name)
                cur_dict[k] = {"value": value_type(value) if value_type else value}

    return params_dict


def str2bool(val: Union[str, bool]) -> bool:
    """If input type is string, convert it to boolean.

    Args:
        val (Union[str, bool]): value to convert to boolean.

    Raises:
        argparse.ArgumentTypeError: If type is neither string and boolean, raise an error.

    Returns:
        bool: return converted boolean value.
    """
    if isinstance(val, bool):
        return val
    if isinstance(val, str):
        if val.lower() in ("true", "1"):
            return True
        if val.lower() in ("false", "0"):
            return False
    raise argparse.ArgumentTypeError("Boolean value expected.")


class ShortDefaultsHelpFormatter(argparse.RawTextHelpFormatter):
    """Text Help Formatter that shortens."""

    def _get_default_metavar_for_optional(self, action):
        return action.dest.split(".")[-1].upper()


def add_hyper_parameters_sub_parser(
    parser, config, modes=None, return_sub_parser=False
) -> Optional[argparse.ArgumentParser]:
    """Adds hyper parameters sub parser."""

    default_modes = ("TRAINING", "INFERENCE")
    if modes is None:
        modes = default_modes
    assert isinstance(modes, tuple)
    for mode in modes:
        assert mode in default_modes

    params = gen_param_help(config)

    subparsers = parser.add_subparsers(help="sub-command help")
    parser_a = subparsers.add_parser(
        "params",
        help="Hyper parameters defined in template file.",
        formatter_class=ShortDefaultsHelpFormatter,
    )
    for k, val in params.items():
        param_type = val["type"]
        if val["affects_outcome_of"] not in modes:
            continue
        if param_type == bool:
            param_type = str2bool
        parser_a.add_argument(
            f"--{k}",
            default=val["default"],
            help=val["help"],
            dest=f"params.{k}",
            type=param_type,
        )
    if return_sub_parser:
        return parser_a
    return None


def get_parser_and_hprams_data():
    """A function to distinguish between when there is template input and when there is no template input.

    Inspect the template using pre_parser to get the template's hyper_parameters information.
    Finally, it returns the parser used in the actual main.
    """
    # TODO: Declaring pre_parser to get the template
    pre_parser = argparse.ArgumentParser(add_help=False)
    pre_parser.add_argument("template", nargs="?", default=None)
    parsed, _ = pre_parser.parse_known_args()
    params = []
    if "params" in sys.argv:
        params = sys.argv[sys.argv.index("params") :]

    template = parsed.template
    hyper_parameters = {}
    parser = argparse.ArgumentParser(formatter_class=RawTextHelpFormatter)
    template_config = find_and_parse_model_template(template)
    template_help_str = (
        "Enter the path or ID or name of the template file. \n"
        "This can be omitted if you have train-data-roots or run inside a workspace."
    )

    if isinstance(template_config, ModelTemplate):
        sys.argv[sys.argv.index(template)] = template_config.model_template_path
        hyper_parameters = template_config.hyper_parameters.data
        parser.add_argument("template", help=template_help_str)
    elif Path("./template.yaml").exists():
        # Workspace Environments
        template_config = parse_model_template("./template.yaml")
        hyper_parameters = template_config.hyper_parameters.data
        parser.add_argument("template", nargs="?", default="./template.yaml", help=template_help_str)
    else:
        parser.add_argument("template", nargs="?", default=None, help=template_help_str)

    return parser, hyper_parameters, params



"""Tests whether the dataloaders can load the data correctly."""

# Copyright (C) 2021-2023 Intel Corporation
# SPDX-License-Identifier: Apache-2.0

import pytest

from otx.api.entities.model_template import TaskType
from tests.unit.algorithms.anomaly.helpers.dummy_dataset import ShapesDataModule


@pytest.mark.parametrize("stage", ["predict", "fit", "validate", "test"])
@pytest.mark.parametrize("task_type", [TaskType.ANOMALY_CLASSIFICATION, TaskType.ANOMALY_DETECTION])
def test_dataloaders(task_type, stage):
    """Tests whether the datamodule can load the data correctly.

    For all the test stages and the task types, the datamodule should return the correct keys.
    """
    datamodule = ShapesDataModule(task_type)
    datamodule.setup(stage)
    if stage == "fit":
        batch = next(iter(datamodule.train_dataloader()))
    elif stage == "validate":
        batch = next(iter(datamodule.val_dataloader()))
    elif stage == "test":
        batch = next(iter(datamodule.test_dataloader()))
    else:
        batch = next(iter(datamodule.predict_dataloader()))
    if task_type == TaskType.ANOMALY_CLASSIFICATION:
        assert batch.keys() == {"image", "label", "index"}
    elif task_type == TaskType.ANOMALY_DETECTION:
        assert batch.keys() == {"image", "label", "index", "boxes"}
    else:
        assert batch.keys() == {"image", "label", "index", "mask"}
